from quantplay.service import market
from uuid import uuid4
from quantplay.broker.generics.broker import Broker

class BrokerClient(Broker):
    def __init__(self):
        self.market_data = {}
        self.current_tick = None
        self.client_orders = {}
        self.client_positions = {}

    def get_holding(self):
        return {}

    def positions(self):
        return self.client_positions

    def orders(self, tag=None):
        orders = list(self.client_orders.values())

        if tag:
            orders = [a for a in orders if a['tag'] == tag]

        return orders

    def ping(self, timestamp):
        self.current_tick = timestamp

        overall_pnl = 0
        for tradingsymbol in self.client_positions:
            ltp = self.ltp(tradingsymbol)
            high = self.tick_high(tradingsymbol)

            position_data = self.client_positions[tradingsymbol]
            quantity = position_data['quantity']
            avg_price = position_data['avg_price']

            self.client_positions[tradingsymbol]['ltp'] = ltp
            self.client_positions[tradingsymbol]['pnl'] = quantity*(ltp*100-avg_price*100)
            overall_pnl += self.client_positions[tradingsymbol]['pnl']

        print("overall pnl at {} is {}".format(timestamp, overall_pnl))

    def load_data(self, symbol):
        if symbol in self.market_data:
            return
        data = market.data(interval="minute", symbols_by_security_type={"OPT": [symbol]})
        self.market_data[symbol] = data.set_index(["symbol", "date"])

    def ltp(self, tradingsymbol):
        return self.market_data[tradingsymbol].loc[(tradingsymbol, self.current_tick)]['close']

    def tick_high(self, tradingsymbol):
        return self.market_data[tradingsymbol].loc[(tradingsymbol, self.current_tick)]['high']

    def place_order(self, tradingsymbol=None, exchange=None, quantity=None, order_type=None, transaction_type=None,
                    tag=None, product=None, price=None, trigger_price=None):
        ltp = self.ltp(tradingsymbol)

        if order_type == 'MARKET':
            if tradingsymbol in self.client_positions:
                raise Exception("Adding position on symbol {} is not allowed".format(tradingsymbol))

            if transaction_type == 'SELL':
                quantity = - quantity

            self.client_positions[tradingsymbol] = {
                'quantity' : quantity,
                'avg_price' : ltp,
                'exchange' : exchange,
                'product' : product
            }
        elif order_type == "SL":
            if transaction_type == "BUY":
                assert trigger_price is not None
                self.client_orders[str(uuid4())] = {
                    'symbol' : tradingsymbol,
                    'quantity' : quantity,
                    'type' : transaction_type,
                    'trigger_price' : trigger_price,
                    'price' : price,
                    'tag' : tag,
                    'side' : transaction_type
                }
            else:
                raise Exception("{} not supported for SL orders")
        else:
            raise Exception("order type {} not supported".format(order_type))

    def execute_order(self, tradingsymbol=None, exchange=None, quantity=None, order_type=None, transaction_type=None,
                      stoploss=None, tag=None, product=None):
        self.load_data(tradingsymbol)
        ltp = self.ltp(tradingsymbol)
        super().execute_order(tradingsymbol=tradingsymbol, exchange=exchange, quantity=quantity, order_type=order_type,
                              transaction_type=transaction_type, stoploss=stoploss, tag=tag, product=product,
                              price=ltp)





import unittest

from quantplay.executor.strategy_executor import UserStrategiesExecutor


class QuantplayStrategyExecutorValidation(unittest.TestCase):
    def test_go_live(self):
        UserStrategiesExecutor("Zerodha", "quantplay/strategy", ["Musk"])


if __name__ == "__main__":
    unittest.main()



from retrying import retry
from selenium import webdriver

from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support.ui import WebDriverWait

class Selenium():
    def __init__(self):
        self.x = False

    @staticmethod
    @retry(wait_exponential_multiplier=1000, wait_exponential_max=10000, stop_max_attempt_number=3)
    def get_browser(headless=True):
        opts = webdriver.ChromeOptions()

        if headless == True:
            opts.add_argument('headless')

        opts.add_argument('no-sandbox')
        browser = webdriver.Chrome(chrome_options=opts)

        return browser

    @staticmethod
    def get_element(browser, element_xpath, delay=10):
        if delay == None:
            return browser.find_element("xpath", element_xpath)
        try:
            element = WebDriverWait(browser, delay).until(
                EC.presence_of_element_located((By.XPATH, element_xpath)))
            return element
        except TimeoutException:
            print("Loading took too much time!")
            raise Exception("Timeout in first load")


class InvalidArgumentException(Exception):
    code = "400"

    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(InvalidArgumentException, self).__init__(message)


class DataNotFoundException(Exception):
    code = "404"

    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(DataNotFoundException, self).__init__(message)


class AccessDeniedException(Exception):
    code = "403"

    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(AccessDeniedException, self).__init__(message)


class RetryableException(Exception):
    code = "409"

    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(RetryableException, self).__init__(message)


def retry_on_access_denied(exc):
    return isinstance(exc, AccessDeniedException)


def retry_exception(exc):
    return isinstance(exc, RetryableException)


class StaleDataFound(Exception):
    code = "101"

    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(StaleDataFound, self).__init__(message)


class ServiceException(Exception):
    code = "500"

    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(ServiceException, self).__init__(message)


class StockNotListedOnExchange(Exception):
    stocks = []

    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(StockNotListedOnExchange, self).__init__(message)


class QuantplayOrderPlacementException(Exception):
    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(QuantplayOrderPlacementException, self).__init__(message)


class StrategyInvocationException(Exception):
    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(StrategyInvocationException, self).__init__(message)



import numpy as np
import pandas as pd
import copy
from datetime import timedelta

from quantplay.reporting.strategy_report import StrategyReport

from quantplay.utils.constant import (
    TickInterval,
    StrategyType,
)
from quantplay.utils.constant import Constants


class Backtesting:
    def __init__(self, market):
        self.market_data = []
        self.market = market

    def load_data(self, symbols_by_security_type, interval):
        self.market_data = self.market.data(
            interval=interval,
            symbols_by_security_type=symbols_by_security_type,
        )

        self.market_data.loc[:, "tradingsymbol"] = self.market_data.symbol
        self.market_data.loc[:, "date_only"] = self.market_data.date.dt.date
        self.market_data.loc[:, "date_only"] = pd.to_datetime(self.market_data.date_only)

        self.business_days_df = (
            self.market_data[["date_only"]]
            .sort_values(["date_only"])
            .groupby(["date_only"])
            .first()
            .reset_index()
        )

    def add_time(self, trades, column_time_tuple=None, holding_days=0):
        if not column_time_tuple:
            raise Exception(
                "column_time_tuple should be like ('column_name', 'hh:mm') e.g. ('exit_time', '15:29')"
            )
        time_column, time_value = column_time_tuple
        hour, minute = [int(a) for a in time_value.split(":")]

        if holding_days > 0:
            self.business_days_df.loc[
                :, "next_business_date_only"
            ] = self.business_days_df.date_only.shift(-holding_days)

            trades = pd.merge(
                trades,
                self.business_days_df,
                how="left",
                left_on=["date_only"],
                right_on=["date_only"],
            )
            trades.loc[:, time_column] = trades.next_business_date_only

            trades.loc[:, time_column] = trades[time_column].apply(
                lambda x: x.replace(hour=hour, minute=minute)
            )
        else:
            trades.loc[:, time_column] = trades.date.apply(
                lambda x: x.replace(hour=hour, minute=minute)
            )

        return trades[~trades[time_column].isna()]

    def add_intraday_metrics(self, entry_time, exit_time):
        self.market_data.loc[:, "price_column"] = self.market_data.close

        condition = (
            self.market_data.date.dt.time <= pd.to_datetime(entry_time).time()
        ) | (self.market_data.date.dt.time > pd.to_datetime(exit_time).time())
        self.market_data.loc[:, "price_column"] = np.where(
            condition, np.nan, self.market_data.price_column
        )

        self.market_data.loc[:, "intraday_high"] = self.market_data.groupby(
            ["symbol", "date_only"]
        ).price_column.transform(max)
        self.market_data.loc[:, "intraday_low"] = self.market_data.groupby(
            ["symbol", "date_only"]
        ).price_column.transform(min)

    def merge_columns(self, trades, columns=[], time_column_name=None):
        for column in columns:
            if column in trades.columns:
                trades = trades.drop([column], axis=1)

        self.market_data.loc[:, time_column_name] = self.market_data.date
        trades = pd.merge(
            trades,
            self.market_data[[time_column_name, "tradingsymbol"] + columns],
            how="left",
            left_on=["tradingsymbol", time_column_name],
            right_on=["tradingsymbol", time_column_name],
        )
        return trades

    def validate_trades(
        self,
        trades,
        columns_to_validate=[
            "transaction_type",
            "strategy_type",
            "exit_time",
            "tag",
            "uuid",
        ],
    ):

        if "transaction_type" in columns_to_validate:
            transaction_types = list(trades.transaction_type.unique())
            assert len(transaction_types) <= 2
            assert set(transaction_types).issubset({"BUY", "SELL"})

        if "strategy_type" in columns_to_validate:
            strategy_types = list(trades.strategy_type.unique())
            assert len(strategy_types) == 1
            assert strategy_types[0] in [
                StrategyType.intraday,
                StrategyType.overnight,
            ]

        if "exit_time" in columns_to_validate:
            assert "exit_time" in trades.columns, "Exit time must be provided in trades"

        if "tag" in columns_to_validate:
            assert "tag" in trades.columns

        if "uuid" in columns_to_validate:
            assert "uuid" in trades.columns

    def backtest(self, trades, interval="minute"):
        security_types = list(trades.security_type.unique())
        symbols_by_security_type = {}
        for security_type in security_types:
            symbols_by_security_type[security_type] = list(
                trades[trades.security_type == security_type].tradingsymbol.unique()
            )

        trades.loc[:, "date_only"] = pd.to_datetime(trades.date.dt.date)

        self.load_data(symbols_by_security_type, interval)

        return self.evaluate_performance(
            trades, interval, tag="testing"
        )

    def evaluate_performance(self, trades_df, interval, tag=None, args={}):
        """
        input : pandas dataframe
        columns : tradingsymbol, trade_time, stoploss, exit_time
        """
        trades = copy.deepcopy(trades_df)
        trades.loc[:, "original_date"] = trades.date

        self.market_data.loc[:, "entry_price"] = self.market_data.close
        trades = self.merge_columns(
            trades, columns=["entry_price"], time_column_name="entry_time"
        )

        self.market_data.loc[:, "close_price"] = self.market_data.close
        trades = self.merge_columns(
            trades, columns=["close_price"], time_column_name="exit_time"
        )

        Constants.logger.info(
            "Dropping {} trades because entry price is not available".format(
                trades[trades.entry_price.isna()][["tradingsymbol", "date"]]
            )
        )
        trades = trades[trades.entry_price > 0]
        trades.loc[:, "entry_price"] = trades.entry_price.astype(float)
        Constants.logger.info(
            "Dropping {} trades because close price is not available".format(
                trades[trades.close_price.isna()][["tradingsymbol", "date"]]
            )
        )
        trades = trades[trades.close_price > 0]
        trades.loc[:, "close_price"] = trades.close_price.astype(float)

        strategy_type = list(trades.strategy_type.unique())[0]
        if strategy_type == StrategyType.intraday:
            columns_to_drop = list(
                {"date", "open", "high", "low", "close"}.intersection(
                    set(list(trades.columns))
                )
            )
            if columns_to_drop:
                trades = trades.drop(columns=columns_to_drop)

            trades = pd.merge(
                trades,
                self.market_data[
                    ["tradingsymbol", "date_only", "date", "high", "low", "close"]
                ],
                how="left",
                left_on=["tradingsymbol", "date_only"],
                right_on=["tradingsymbol", "date_only"],
            )

            Backtesting.add_attributes(trades)
            trades = trades[
                (trades.date > trades.entry_time) & (trades.date <= trades.exit_time)
            ]

            trades.loc[:, "is_trade_closed"] = np.where(
                (trades.exit_time == trades.date), 1, 0
            )

            Backtesting.apply_stoploss(trades)
            Backtesting.custom_exit(trades)
            Backtesting.apply_trailing_stoploss(trades)
            Backtesting.apply_trade_squareoff(trades)

            trades = trades[trades.is_trade_closed == 1]

            trades = trades.groupby(["tradingsymbol", "entry_time"]).head(1).reset_index()
            trades.loc[:, "closing_timestamp"] = trades.date
        else:
            trades.loc[:, "closing_timestamp"] = trades.exit_time

        interval_minutes = int(TickInterval.get_interval_seconds(interval) / 60)
        trades.loc[:, "order_timestamp"] = trades.entry_time + timedelta(
            minutes=interval_minutes
        )
        trades.loc[:, "date"] = trades.original_date
        trades.loc[:, "exposure"] = trades.quantity * trades.entry_price

        fno_slippage = 0
        if 'fno_slippage' in args:
            fno_slippage = args['fno_slippage']

        cm_slippage = 0
        if 'cm_slippage' in args:
            cm_slippage = args['cm_slippage']

        response = StrategyReport.generate_report(trades, fno_slippage=fno_slippage, cm_slippage=cm_slippage)
        result = pd.DataFrame(response)

        trades = StrategyReport.add_more_params(trades,
                                                fno_slippage=fno_slippage,
                                                cm_slippage=cm_slippage)

        StrategyReport.print_report(result)
        if "display_visuals" in args and args["display_visuals"] == False:
            Constants.logger.info("Skipping report display")
        else:
            StrategyReport.print_portfolio(trades, tag)

        return result, trades

    @staticmethod
    def apply_trade_squareoff(trades):

        if "trade_squareoff" not in trades.columns:
            return

        print("Applying trade squareoff")

        trades.loc[:, "trade_squareoff_price"] = np.where(
            trades.transaction_type == "SELL",
            Constants.round_to_tick((1 - trades.trade_squareoff) * trades.entry_price),
            Constants.round_to_tick((1 + trades.trade_squareoff) * trades.entry_price),
        )

        trades.loc[:, "is_trade_closed"] = np.where(
            (
                (
                    (trades.transaction_type == "SELL")
                    & (trades.low <= trades.trade_squareoff_price)
                )
                | (
                    (trades.transaction_type == "BUY")
                    & (trades.high >= trades.trade_squareoff_price)
                )
            ),
            1,
            trades.is_trade_closed,
        )

        trades.loc[:, "close_price"] = np.where(
            (
                (
                    (trades.transaction_type == "SELL")
                    & (trades.low <= trades.trade_squareoff_price)
                )
                | (
                    (trades.transaction_type == "BUY")
                    & (trades.high >= trades.trade_squareoff_price)
                )
            ),
            trades.trade_squareoff_price,
            trades.close_price,
        )

    @staticmethod
    def apply_strategy_squareoff(trades):

        if "strategy_squareoff" not in trades.columns:
            return trades

        print("Applying strategy squareoff")

        trades.loc[:, "squareoff_profit"] = trades.strategy_squareoff * trades.entry_price
        trades.loc[:, "trade_profit"] = np.where(
            trades.transaction_type == "SELL",
            trades.entry_price - trades.close,
            trades.close - trades.entry_price,
        )

        trades_combined_profits = (
            trades.groupby(["date", "uuid"])["squareoff_profit", "trade_profit"]
            .sum()
            .reset_index()
        )

        trades = trades.drop(columns=["squareoff_profit", "trade_profit"])

        trades = pd.merge(
            trades,
            trades_combined_profits,
            how="left",
            left_on=["date", "uuid"],
            right_on=["date", "uuid"],
        )

        trades.loc[:, "is_trade_closed"] = np.where(
            trades.trade_profit >= trades.squareoff_profit,
            1,
            trades.is_trade_closed,
        )

        trades.loc[:, "close_price"] = np.where(
            trades.trade_profit >= trades.squareoff_profit,
            trades.close,
            trades.close_price,
        )

        return trades

    @staticmethod
    def apply_stoploss(trades):

        if "stoploss" not in trades.columns:
            return
        print("Applying stoploss")

        trades.loc[:, "stoploss_price"] = np.where(
            trades.transaction_type == "SELL",
            Constants.round_to_tick((1 + trades.stoploss) * trades.entry_price),
            Constants.round_to_tick((1 - trades.stoploss) * trades.entry_price),
        )

        trades.loc[:, "is_trade_closed"] = np.where(
            (
                (
                    (trades.transaction_type == "SELL")
                    & (trades.high >= trades.stoploss_price)
                )
                | (
                    (trades.transaction_type == "BUY")
                    & (trades.low <= trades.stoploss_price)
                )
            ),
            1,
            trades.is_trade_closed,
        )
        trades.loc[:, "close_price"] = np.where(
            (
                (
                    (trades.transaction_type == "SELL")
                    & (trades.high >= trades.stoploss_price)
                )
                | (
                    (trades.transaction_type == "BUY")
                    & (trades.low <= trades.stoploss_price)
                )
            ),
            trades.stoploss_price,
            trades.close_price,
        )

    @staticmethod
    def apply_trailing_stoploss(trades):

        if "trailing_stoploss" not in trades.columns:
            return
        print("Applying trailing stoploss")

        trades.loc[:, "entry_sl_trigger_price"] = np.where(
            trades.transaction_type == "SELL",
            Constants.round_to_tick((1 + trades.trailing_stoploss) * trades.entry_price),
            Constants.round_to_tick((1 - trades.trailing_stoploss) * trades.entry_price),
        )
        trades.loc[:, "sl_trigger_price"] = np.where(
            trades.transaction_type == "SELL",
            Constants.round_to_tick((1 + trades.trailing_stoploss) * trades.close),
            Constants.round_to_tick((1 - trades.trailing_stoploss) * trades.close),
        )

        trades.loc[:, "sl_trigger_price"] = np.minimum(trades.sl_trigger_price, trades.entry_sl_trigger_price)

        trades.loc[:, "is_trade_closed"] = np.where(
            (
                    (
                            (trades.transaction_type == "SELL")
                            & (trades.high >= trades.sl_trigger_price)
                    )
                    | (
                            (trades.transaction_type == "BUY")
                            & (trades.low <= trades.sl_trigger_price)
                    )
            ),
            1,
            trades.is_trade_closed,
        )
        trades.loc[:, "close_price"] = np.where(
            (
                    (
                            (trades.transaction_type == "SELL")
                            & (trades.high >= trades.sl_trigger_price)
                    )
                    | (
                            (trades.transaction_type == "BUY")
                            & (trades.low <= trades.sl_trigger_price)
                    )
            ),
            trades.sl_trigger_price,
            trades.close_price,
        )

    @staticmethod
    def custom_exit(trades):

        if "custom_exit" not in trades.columns:
            return

        print("Applying custom exit")

        assert len(trades.transaction_type.unique()) == 1
        assert trades.transaction_type.unique()[0] == "SELL"

        trades.loc[:, "is_trade_closed"] = np.where(trades.close >= trades.sma_9,
                                                    1, trades.is_trade_closed)
        trades.loc[:, "close_price"] = trades.close

    @staticmethod
    def add_attributes(trades):

        if "custom_exit" not in trades.columns:
            return

        trades.loc[:, "sma_9"] = trades.close.rolling(9).mean()


from quantplay.service import market
import pandas as pd
import time
from quantplay.broker.client import broker_client

backtest_date = "2022-01-20"

def load_data(data_date):
    # Load data
    data = market.data(interval="minute", symbols_by_security_type={"EQ": ["NIFTY BANK"]})
    data = data[data.date > "{} 08:15:00".format(data_date)]
    data = data[data.date < "{} 15:15:00".format(data_date)]
    data.loc[:, "day_of_week"] = data.date.dt.day_name()
    data = data[data.date.dt.hour < 17]

    # Add additional attributes
    data = market.add_expiry(data, security_type="OPT", days_offset=0)

    return data

data = load_data(backtest_date)

data_seq = data.to_dict('records')

for d in data_seq:
    timestamp = d['date']
    broker_client.ping(timestamp)
    ltp = d['open']
    strike_gap = d['strike_gap']
    expiry = d['expiry_date']
    if timestamp.hour == 9 and timestamp.minute < 30:
        print("waiting for market to cool down")
        continue

    if timestamp.hour > 13:
        print("market is about to close, let's not place any orders")
        continue

    orders = broker_client.orders(tag="ladder")

    if len(orders) == 0:
        atm = int(round(ltp / strike_gap) * strike_gap)
        pe_symbol = broker_client.option_symbol("NIFTY BANK", expiry, atm, "PE")
        ce_symbol = broker_client.option_symbol("NIFTY BANK", expiry, atm, "CE")

        broker_client.execute_order(
            tradingsymbol=pe_symbol,
            exchange="NSE",
            quantity=100,
            order_type='MARKET',
            transaction_type='SELL',
            stoploss=0.8,
            tag="ladder",
            product="NRML"
        )
        broker_client.execute_order(
            tradingsymbol=ce_symbol,
            exchange="NSE",
            quantity=100,
            order_type='MARKET',
            transaction_type='SELL',
            stoploss=0.8,
            tag="ladder",
            product="NRML"
        )


import copy

import numpy as np
import pandas as pd
from quantplay.utils.transaction_utils import TransactionCost
from quantplay.reporting.visuals import VisualReport

from quantplay.utils.constant import Order, ExchangeName
from quantplay.utils.constant import (
    StrategyMetricsTableColumns,
)


class StrategyReport:
    @staticmethod
    def max_drawdown(vec):
        vec = vec.fillna(method="ffill").fillna(method="bfill")
        maximums = np.maximum.accumulate(vec)
        drawdowns = maximums - vec
        #
        # for i in range(0, 140):
        #     print(vec[i], maximums[i], drawdowns[i])

        return np.max(drawdowns)

    @staticmethod
    def max_drawdown_days(vec):
        start_index = 0
        running_max_balance = 0
        max_drawdown_days = 0
        for i in range(0, len(vec)):
            balance = vec[i]

            if balance > running_max_balance:
                running_max_balance = balance
                start_index = i

            max_drawdown_days = max(max_drawdown_days, i - start_index)

        return max_drawdown_days

    @staticmethod
    def get_sharpe_ratio(portfolio):
        portfolio["today_date"] = pd.to_datetime(portfolio.order_timestamp)
        portfolio["today_date"] = portfolio.today_date.dt.date
        t_df = portfolio.groupby("today_date")["exposure"].sum().reset_index()
        exposure_90 = t_df.exposure.quantile(0.9)

        portfolio["Daily_Returns"] = portfolio.profit / portfolio.exposure

        Sharpe_Ratio = (
            portfolio["Daily_Returns"].mean() / portfolio["Daily_Returns"].std()
        )
        Sharpe_Ratio = Sharpe_Ratio * 100

        portfolio["Date"] = portfolio["order_timestamp"] - pd.to_timedelta(7, unit="d")
        data = (
            portfolio.groupby([pd.Grouper(key="Date", freq="W-MON")])[
                ["profit", "exposure"]
            ]
            .sum()
            .reset_index()
            .sort_values("Date")
        )
        data["Daily_Returns"] = data["profit"] / abs(data["exposure"])
        weekly_sharpe_ratio = data["Daily_Returns"].mean() / data["Daily_Returns"].std()
        weekly_sharpe_ratio = weekly_sharpe_ratio * 100

        portfolio["Date"] = portfolio["order_timestamp"]
        monthly_data = (
            portfolio.groupby([pd.Grouper(key="Date", freq="M")])[["profit", "exposure"]]
            .sum()
            .reset_index()
            .sort_values("Date")
        )
        monthly_data["Daily_Returns"] = monthly_data["profit"] / abs(
            monthly_data["exposure"]
        )
        monthly_sharpe_ratio = (
            monthly_data["Daily_Returns"].mean() / monthly_data["Daily_Returns"].std()
        )
        monthly_sharpe_ratio = monthly_sharpe_ratio * 100

        response = {}
        response[StrategyMetricsTableColumns.weekly_sharpe_ratio] = weekly_sharpe_ratio
        response[StrategyMetricsTableColumns.monthly_sharpe_ratio] = monthly_sharpe_ratio
        response[StrategyMetricsTableColumns.sharpe_ratio] = Sharpe_Ratio
        response[StrategyMetricsTableColumns.monthly_pnl] = monthly_data["profit"].mean()
        response[StrategyMetricsTableColumns.exposure_90] = exposure_90

        return response

    @staticmethod
    def print_report(response):
        result = pd.DataFrame(response)
        print(
            result[
                [
                    "year",
                    "bps",
                    "monthly_pnl",
                    "max_drawdown_days",
                    "max_drawdown",
                    "sharpe_ratio",
                    "total_signals",
                    "success_ratio",
                    "exposure_90",
                    "total_profit",
                    "unique_stocks",
                ]
            ]
        )

    @staticmethod
    def print_portfolio(portfolio, tag=None):
        file_prefix = ""
        if tag is not None:
            file_prefix = "{}_".format(tag)

        VisualReport.display_profit_report(
            portfolio, groupby="M", file_prefix=file_prefix
        )
        VisualReport.display_profit_report(
            portfolio, groupby="Y", file_prefix=file_prefix
        )
        VisualReport.display_balance_report(portfolio, file_prefix=file_prefix)

    @staticmethod
    def add_more_params(portfolio, cm_slippage=0, fno_slippage=0):
        # Basic trade checks
        portfolio = portfolio[portfolio.close_price > 0]
        portfolio = portfolio[portfolio.entry_price > 0]

        portfolio.loc[:, "order_timestamp"] = pd.to_datetime(portfolio["order_timestamp"])
        portfolio.loc[:, "order_timestamp"] = portfolio.sort_values("order_timestamp")

        portfolio.loc[:, "profit"] = portfolio.exposure * (
            1 - portfolio.close_price / portfolio.entry_price
        )

        portfolio.loc[:, Order.profit] = np.where(
            portfolio[Order.transaction_type] == Order.sell_transaction,
            portfolio.profit,
            -portfolio.profit,
        )

        portfolio = TransactionCost.add_transaction_costs(portfolio)
        portfolio.loc[:, Order.profit] = portfolio.profit - portfolio.total_charges
        portfolio.loc[:, Order.profit] = np.where(
            portfolio.exchange == ExchangeName.nse,
            portfolio.profit - portfolio.exposure * cm_slippage,
            portfolio.profit - portfolio.exposure * fno_slippage,
        )

        portfolio.loc[:, "pct_profit"] = portfolio.profit / portfolio.exposure

        portfolio.loc[:, "day_of_week"] = portfolio.order_timestamp.dt.day_name()
        portfolio.loc[:, "point_diff"] = np.where(
            portfolio[Order.transaction_type] == Order.sell_transaction,
            portfolio.entry_price - portfolio.close_price,
            portfolio.close_price - portfolio.entry_price,
        )

        portfolio.loc[:, "option_type"] = np.where(
            portfolio["tradingsymbol"].str.contains("PE"), "PE", "CE"
        )

        return portfolio

    @staticmethod
    def generate_report(portfolio, cm_slippage=0, fno_slippage=0):
        portfolio = StrategyReport.add_more_params(portfolio, cm_slippage, fno_slippage)

        portfolio.loc[:, "order_timestamp"] = pd.to_datetime(portfolio["order_timestamp"])
        portfolio = portfolio.sort_values("order_timestamp")

        portfolio.loc[:, "balance"] = portfolio.profit.cumsum()

        portfolio = portfolio.reset_index(drop=True)

        tag = portfolio["tag"].iat[0]

        portfolio["year"] = portfolio.order_timestamp.dt.year
        years = portfolio["year"].unique()

        response = StrategyReport.create_single_report(copy.deepcopy(portfolio))
        response[StrategyMetricsTableColumns.tag] = tag
        response["year"] = "ALL"

        response_list = []
        response_list.append(copy.deepcopy(response))

        for i in range(0, len(years)):
            year = int(years[i])
            data = portfolio[portfolio.year == year]
            response = StrategyReport.create_single_report(data)
            response[StrategyMetricsTableColumns.tag] = tag
            response["year"] = year
            response_list.append(copy.deepcopy(response))
        return response_list

    @staticmethod
    def create_report(strategy_params):
        input = strategy_params
        strategy_params = {}
        strategy_params[StrategyMetricsTableColumns.tag] = input["tag"]
        portfolio = pd.read_csv("/tmp/{}.csv".format(strategy_params["tag"]))

        return StrategyReport.generate_report(portfolio)

    @staticmethod
    def create_single_report(portfolio):
        response = {}
        portfolio = portfolio.sort_values("order_timestamp")

        response["total_profit"] = portfolio["profit"].sum()

        portfolio.loc[:, "balance"] = portfolio.profit.cumsum()

        portfolio.loc[:, "date"] = portfolio.order_timestamp.dt.date

        # unique stock count
        t_df = portfolio.groupby("date").symbol.nunique().reset_index()
        response[StrategyMetricsTableColumns.unique_stocks] = int(t_df.symbol.mean())

        day_wise_pnl = portfolio.groupby(["date"])["profit"].sum().reset_index()
        day_wise_pnl = day_wise_pnl.sort_values("date")

        day_wise_pnl.loc[:, "is_loss"] = np.where(day_wise_pnl.profit > 0, 0, 1)
        m = day_wise_pnl.is_loss.astype(bool)
        day_wise_pnl.loc[:, "loss_streak"] = (
            m.groupby([m, (~m).cumsum().where(m)]).cumcount().add(1).mul(m)
        )

        day_wise_pnl.loc[:, "is_win"] = np.where(day_wise_pnl.profit > 0, 1, 0)
        m = day_wise_pnl.is_win.astype(bool)
        day_wise_pnl.loc[:, "win_streak"] = (
            m.groupby([m, (~m).cumsum().where(m)]).cumcount().add(1).mul(m)
        )

        day_wise_pnl.loc[:, "balance"] = day_wise_pnl.profit.cumsum()
        max_drawdown = StrategyReport.max_drawdown(day_wise_pnl["balance"])
        max_drawdown_days = StrategyReport.max_drawdown_days(day_wise_pnl["balance"])

        sharpe_response = StrategyReport.get_sharpe_ratio(portfolio)

        response[StrategyMetricsTableColumns.max_drawdown] = "{}".format(max_drawdown)
        response[StrategyMetricsTableColumns.max_drawdown_days] = "{}".format(
            max_drawdown_days
        )
        response[StrategyMetricsTableColumns.max_gain] = portfolio["profit"].max()
        response[StrategyMetricsTableColumns.max_loss] = portfolio["profit"].min()
        response[StrategyMetricsTableColumns.lossing_streak] = int(
            day_wise_pnl.loss_streak.max()
        )
        response[StrategyMetricsTableColumns.winning_streak] = int(
            day_wise_pnl.win_streak.max()
        )

        for key in sharpe_response:
            response[key] = sharpe_response[key]

        profit_trade = portfolio[portfolio.profit > 0]
        lossing_trade = portfolio[portfolio.profit < 0]

        response[StrategyMetricsTableColumns.avg_profit] = profit_trade["profit"].mean()
        response[StrategyMetricsTableColumns.avg_loss] = lossing_trade["profit"].mean()

        response["number_of_losses"] = len(lossing_trade)
        response["number_of_wins"] = len(profit_trade)
        response["total_signals"] = len(portfolio)
        response["total_trading_days"] = len(portfolio.order_timestamp.dt.date.unique())
        response["drawdown_pnl_ratio"] = (
            float(response["max_drawdown"]) / response["monthly_pnl"]
        )
        response["success_ratio"] = response["number_of_wins"] / (
            response["number_of_losses"] + response["number_of_wins"]
        )
        response["avg_pnl"] = portfolio["profit"].mean()

        response["bps"] = (
            portfolio["profit"].sum() / portfolio["exposure"].sum()
        ) * 10000

        return response


if __name__ == "__main__":
    strategy_params = {"tag": "matsya"}

    pass



import threading
# TODO: Import other broker classes
import traceback
from collections import defaultdict
from queue import Empty as empty_exception
from quantplay.utils.constant import Constants

from quantplay.brokerage.zerodha.ZBroker import ZBroker
from quantplay.brokerage.angelone.angel_broker import AngelBroker
from quantplay.exception.exceptions import (
    QuantplayOrderPlacementException,
    StrategyInvocationException,
)
from quantplay.utils.exchange import Market


class UserStrategiesExecutor:
    def __init__(self, broker_name, strategies):
        self.strategies = strategies

        for s in self.strategies:
            s.validate_production()
            s.add_derivative_symbols(mode='prod')

        required_tick_intervals = list(
            set([s.required_tick_interval() for s in self.strategies])
        )

        symbols_to_subscribe = defaultdict(set)
        for s in self.strategies:
            for (
                security_type,
                symbols,
            ) in s.stream_symbols_by_security_type.items():
                stream_symbols_from_exchange = Market.EXCHANGE_MAPPINGS[
                    s.exchange_to_trade_on
                ][security_type]
                symbols_to_subscribe[stream_symbols_from_exchange].update(symbols)

        self.broker = None
        if broker_name == "Zerodha":
            self.broker = ZBroker(required_tick_intervals, True)
        elif broker_name == "AngelOne":
            print(required_tick_intervals)
            self.broker = AngelBroker(required_tick_intervals, True)
        else:
            raise Exception(
                f"Broker {self.broker} not supported yet. Supported Brokers: Zerodha"
            )

        for exchange, symbols in symbols_to_subscribe.items():
            self.broker.add_strategy_symbols(list(symbols), exchange)

        Constants.logger.info(f"Strategies Instruments Subscribed: {self.broker.strategy_instruments}")

        self.broker.bootstrap_strategy_child_orders_map(self.strategies)

        load_data_for_days = max([s.data_required_for_days for s in self.strategies])

        Constants.logger.info(f"Loading Historical Data for {load_data_for_days} days")

        self.broker.load_historical_market_feed(load_data_for_days)

        Constants.logger.info(f"Historical data loaded")

    def run_strategies(self):
        while True:
            try:
                (
                    interval,
                    tick_time,
                ) = self.broker.processed_tick_interval_queue.get(timeout=5)
                Constants.logger.info("Processing {} {}".format(interval, tick_time))
                print("Processing {} {}".format(interval, tick_time))
                for strategy in self.strategies:
                    if strategy.should_exit(tick_time):
                        Constants.logger.info(
                            f"Closing child orders for strategy {strategy.strategy_tag}"
                        )
                        self.broker.close_child_orders(tag=strategy.strategy_tag)
                        Constants.logger.info(
                            f"Strategy {strategy.strategy_tag} child orders are modified to be executed immediately"
                        )

                    if strategy.should_invoke(interval, tick_time):
                        Constants.logger.info(f"Invoking Strategy {strategy}")
                        print(f"Invoking Strategy {strategy}")

                        market_data = self.broker.market_df(
                            strategy.required_tick_interval()
                        )

                        market_data = market_data[
                            market_data.security_type.isin(["OPT", "FUT"])
                            | market_data.symbol.isin(
                                strategy.stream_symbols_by_security_type["EQ"]
                            )
                        ]

                        try:
                            orders = strategy.live_orders(market_data, tick_time)
                        except StrategyInvocationException as e:
                            Constants.logger.info(
                                f"[STRATEGY_INVOCATION_FAILED] Failed to invoke {strategy.strategy_tag} got {e}"
                            )
                            traceback.print_exc()
                            continue

                        if not len(orders):
                            Constants.logger.info(
                                f"No orders to be executed for strategy {strategy.strategy_tag}"
                            )
                            continue



                        for order in orders:
                            Constants.logger.info("order [{}] by [{}]".format(strategy.strategy_tag, order))
                            # try:
                            #     order_id = self.broker.place_or_modify_order(
                            #         order, is_new_order=True
                            #     )
                            #     print(f"Executed Order {order} OrderId {order_id}")
                            # except QuantplayOrderPlacementException as e:
                            #     print(
                            #         f"[ORDER_PLACEMENT_FAILED] Failed to place {strategy.strategy_tag} order {order} got {e}"
                            #     )
                            #     traceback.print_exc()

                        Constants.logger.info(f"Finished executing strategy {strategy}")
            except empty_exception:
                Constants.logger.info("waiting for new tick addition to market dataframe")
                print("waiting for new tick addition to market dataframe")
                pass
            except Exception as e:
                Constants.logger.info(f"[CRITICAL_UNKNOWN_EXCEPTION] something went wrong {e}")
                traceback.print_exc()

    def start_execution(self):
        th = threading.Thread(target=self.run_strategies)
        th.start()
        self.broker.connect()





from quantplay.utils.constant import TradingCode


class StrategyResponse:
    def __init__(self):
        self.new_orders = {}
        self.trades_to_close = {}
        self.response_code = TradingCode.ok



from quantplay.broker import motilal

motilal = Motilal()
motilal.generate_token()
motilal.send_otp()
# motilal.verify_otp(otp)
ltp = motilal.get_ltp("NSE", "SBIN")
print(ltp)

motilal.place_order("SBIN", "NSE", 1, "LIMIT", "BUY", "test", "NORMAL", ltp, 0)
orders_placed = motilal.get_orders()
print(orders_placed)

motilal.modify_orders_till_complete(orders_placed)


import pandas as pd
import codecs
import pickle
import numpy as np
import time
from retrying import retry
from quantplay.utils.constant import Constants
from quantplay.exception.exceptions import InvalidArgumentException
from quantplay.broker.generics.broker import Broker
from quantplay.config.qplay_config import QplayConfig
from kiteconnect import KiteConnect
import traceback
from quantplay.broker.kite_utils import KiteUtils
from quantplay.utils.exchange import Market as MarketConstants
from datetime import timedelta

class Zerodha(Broker):

    stoploss = 'stoploss'
    zerodha_api_key = "zerodha_api_key"
    zerodha_api_secret = "zerodha_api_secret"
    zerodha_wrapper = "zerodha_wrapper"

    def __init__(self):
        try:
            wrapper = QplayConfig.get_value(Zerodha.zerodha_wrapper)
            self.set_wrapper(wrapper)
            self.wrapper.orders()
        except Exception as e:
            self.wrapper = self.generate_token()
        Constants.logger.info(self.wrapper.profile())

        super(Zerodha, self).__init__()

    def set_wrapper(self, serialized_wrapper):
        self.wrapper = pickle.loads(codecs.decode(serialized_wrapper.encode(), "base64"))

    def set_username(self, username):
        self.username = username

    def get_username(self):
        return self.username

    @retry(wait_exponential_multiplier=3000, wait_exponential_max=10000, stop_max_attempt_number=3)
    def get_ltp(self, exchange=None, tradingsymbol=None):
        try:
            key = "{}:".format(exchange) + tradingsymbol
            response = self.wrapper.ltp([key])

            if key not in response:
                raise InvalidArgumentException("Symbol {} not listed on exchange".format(tradingsymbol))

            response = response[key]['last_price']
            return response
        except Exception as e:
            exception_message = "GetLtp call failed for [{}] with error [{}]".format(tradingsymbol, str(e))
            Constants.logger.error("{}".format(exception_message))

    @retry(wait_exponential_multiplier=3000, wait_exponential_max=10000, stop_max_attempt_number=3)
    def get_orders(self, status=None):
        orders = self.wrapper.orders()
        if status:
            orders = [a for a in orders if a['status'] == status]
        return orders

    def exit_all_trigger_orders(self, tag="ALL", symbol_cotains=None):
        stoploss_orders = self.get_orders("TRIGGER PENDING")

        if len(stoploss_orders) == 0:
            print("All stoploss orders have been already closed")
            return

        stoploss_orders = pd.DataFrame(stoploss_orders)
        if tag != "ALL":
            stoploss_orders = stoploss_orders[stoploss_orders.tag == tag]

        if symbol_cotains is not None:
            stoploss_orders = stoploss_orders[stoploss_orders['tradingsymbol'].str.contains(symbol_cotains)]

        if len(stoploss_orders) == 0:
            print("All stoploss orders have been already closed")
            return

        orders_to_close = list(stoploss_orders.order_id.unique())

        stoploss_orders = stoploss_orders.to_dict('records')
        for stoploss_order in stoploss_orders:
            exchange = stoploss_order['exchange']
            tradingsymbol = stoploss_order['tradingsymbol']

            if exchange == "NFO":
                stoploss_order['order_type'] = "MARKET"
                stoploss_order['price'] = 0
            else:
                ltp = self.get_ltp(exchange, tradingsymbol)
                stoploss_order['order_type'] = "LIMIT"
                stoploss_order['price'] = self.round_to_tick(ltp)

            self.modify_order(stoploss_order)
            time.sleep(.1)

        self.modify_orders_till_complete(orders_to_close)
        print("All order have been closed successfully")

    @retry(wait_exponential_multiplier=3000, wait_exponential_max=10000, stop_max_attempt_number=3)
    def modify_order(self, data):
        try:
            print("Modifying order [{}] new price [{}]".format(data['order_id'], data['price']))
            response = self.wrapper.modify_order(order_id=data['order_id'],
                                              variety=data['variety'],
                                              price=data['price'],
                                              trigger_price=data['trigger_price'],
                                              order_type=data['order_type'])
            return response
        except Exception as e:
            exception_message = "OrderModificationFailed for {} failed with exception {}".format(data['order_id'], e)
            Constants.logger.error("{}".format(exception_message))

    def get_ltp_by_order(self, order):
        exchange = order['exchange']
        tradingsymbol = order['tradingsymbol']

        return self.get_ltp(exchange, tradingsymbol)

    def modify_orders_till_complete(self, orders_placed):
        modification_count = {}
        while 1:
            time.sleep(10)
            orders = pd.DataFrame(self.get_orders())

            orders = orders[orders.order_id.isin(orders_placed)]
            orders = orders[~orders.status.isin(["REJECTED", "CANCELLED", "COMPLETE"])]

            if len(orders) == 0:
                Constants.logger.info("ALL orders have been completed")
                break

            orders = orders.to_dict('records')
            for order in orders:
                order_id = order['order_id']

                ltp = self.get_ltp(order['exchange'], order['tradingsymbol'])
                order['price'] = ltp
                self.modify_order(order)

                if order_id not in modification_count:
                    modification_count[order_id] = 1
                else:
                    modification_count[order_id] += 1

                time.sleep(.1)

                if modification_count[order_id] > 5:
                    order['order_type'] = "MARKET"
                    order['price'] = 0
                    Constants.logger.info("Placing MARKET order [{}]".format(order))
                    self.modify_order(order)

    @retry(wait_exponential_multiplier=3000, wait_exponential_max=10000, stop_max_attempt_number=3)
    def get_positions(self):
        return self.wrapper.positions()

    def positions_pnl(self):
        positions = pd.DataFrame(self.get_positions()['net'])
        print("Total PnL {}".format(positions.pnl.astype(float).sum()))

    def add_params(self, orders):
        df = pd.DataFrame(orders)
        df.loc[:, 'price'] = df.apply(lambda x: self.get_ltp(x['exchange'],
                                                             x['tradingsymbol']),
                                      axis=1)

        df.loc[:, 'disclosedquantity'] = np.where(df.exchange == "NSE", df.quantity/10 + 1, df.quantity)
        df.loc[:, 'disclosedquantity'] = df.disclosedquantity.astype(int)

        return df.to_dict('records')


    # @retry(wait_exponential_multiplier=3000, wait_exponential_max=10000, stop_max_attempt_number=3)
    def place_order(self, tradingsymbol=None, exchange=None, quantity=None, order_type=None, transaction_type=None,
                    tag=None, product=None, price=None, trigger_price=None):
        try:
            order_id = self.wrapper.place_order(variety='regular',
                                                tradingsymbol=tradingsymbol,
                                                exchange=exchange,
                                                transaction_type=transaction_type,
                                                quantity=int(abs(quantity)),
                                                order_type=order_type,
                                                disclosed_quantity=None,
                                                price=price,
                                                trigger_price=trigger_price,
                                                product=product,
                                                tag=tag)
            return order_id
        except Exception as e:
            exception_message = "Order placement failed with error [{}]".format(str(e))
            print(exception_message)

    def configure(self):
        quantplay_config = QplayConfig.get_config()

        print("Enter Zerodha API key:")
        api_key = input()

        print("Enter Zerodha API Secret:")
        api_secret = input()

        quantplay_config['DEFAULT'][Zerodha.zerodha_api_key] = api_key
        quantplay_config['DEFAULT'][Zerodha.zerodha_api_secret] = api_secret

        with open('{}/config'.format(QplayConfig.config_path), 'w') as configfile:
            quantplay_config.write(configfile)

    def validate_config(self, quantplay_config):
        if quantplay_config is None:
            return False
        if Zerodha.zerodha_api_key not in quantplay_config['DEFAULT']:
            return False
        if Zerodha.zerodha_api_secret not in quantplay_config["DEFAULT"]:
            return False

        return True

    def generate_token(self):
        quantplay_config = QplayConfig.get_config()

        if not self.validate_config(quantplay_config):
            self.configure()
            quantplay_config = QplayConfig.get_config()

        api_key = quantplay_config['DEFAULT']['zerodha_api_key']
        api_secret = quantplay_config['DEFAULT']['zerodha_api_secret']
        kite = KiteConnect(api_key=api_key)

        request_token = None
        try:
            request_token = KiteUtils.get_request_token(kite_api_key=api_key)
        except Exception as e:
            traceback.print_exc()
            print("Need token input " + kite.login_url())
            raise e
            # request_token = input()

        print("request token {} api_secret {}".format(request_token, api_secret))

        data = kite.generate_session(request_token, api_secret=api_secret)
        kite.set_access_token(data["access_token"])

        QplayConfig.save_config("zerodha_wrapper", codecs.encode(pickle.dumps(kite), "base64").decode())
        return kite





import re
import numpy as np
import uuid
from quantplay.utils.constant import Constants, OrderType
from quantplay.service import market, backtesting
from quantplay.utils.exchange import Market as MarketConstants
from quantplay.model.exchange.order import QuantplayExchangeOrder
from quantplay.exception.exceptions import StrategyInvocationException
import datetime
import pandas as pd

pd.options.mode.chained_assignment = None  # default='warn'


class QuantplayAlgorithm:
    MAX_LEN_FOR_STRATEGY_TAG = 15

    def __init__(self):
        self.add_derivative_symbols()

    def validate_production(self):
        if not hasattr(self, "strategy_trigger_times"):
            raise Exception("Strategy must have 'strategy_trigger_times' attribute")

        if not hasattr(self, "data_required_for_days"):
            raise Exception("Strategy must have 'data_required_for_days' attribute")

        if not hasattr(self, "execution_algo"):
            raise Exception("Strategy must have 'execution_algo' attribute")

        if not hasattr(self, "order_type"):
            raise Exception("Strategy must have 'order_type' attribute")

        supported_orders = [OrderType.market, OrderType.limit]
        if self.order_type not in supported_orders:
            raise Exception("order type {} not supported, please use {}".format(self.order_type,
                                                                                supported_orders))
        self.validate()

    def validate(self):
        if not hasattr(self, "interval"):
            raise Exception("Strategy must have an 'interval' attribute.")

        if not hasattr(self, "exchange_to_trade_on"):
            raise Exception("Strategy must have an 'exchange_to_trade_on' attribute.")

        if not hasattr(self, "stream_symbols_by_security_type"):
            raise Exception(
                "Strategy must have an 'stream_symbols' dictionary attribute mapping exchange to symbol list"
            )

        if not hasattr(self, "strategy_type"):
            raise Exception(
                "Strategy must have 'strategy_type' attribute which can be either 'intraday' or 'overnight'"
            )

        if not hasattr(self, "entry_time"):
            raise Exception("Strategy must have 'entry_time' attribute")

        if not hasattr(self, "strategy_tag"):
            raise Exception("Strategy must have 'strategy_tag' attribute")
        else:
            if not self.strategy_tag.isalnum():
                raise Exception(
                    "strategy tag {} is not alphanumeric".format(self.strategy_tag)
                )
            if len(self.strategy_tag) > QuantplayAlgorithm.MAX_LEN_FOR_STRATEGY_TAG:
                raise Exception(
                    "length of strategy tag {} must have less than {} characters".format(
                        self.strategy_tag,
                        QuantplayAlgorithm.MAX_LEN_FOR_STRATEGY_TAG,
                    )
                )

        print("Strategy with tag {} successfully validated".format(self.strategy_tag))

    def should_exit(self, tick_time):
        tick_time_str = tick_time.strftime(Constants.hour_min_format)
        x = re.search(self.exit_time, tick_time_str)
        if x:
            return True
        return False

    def should_invoke(self, interval, tick_time):

        if interval != self.interval:
            return False

        tick_time_str = tick_time.strftime(Constants.hour_min_format)

        for trigger_time in self.strategy_trigger_times:
            x = re.search(trigger_time, tick_time_str)
            if x:
                return True
        return False

    def required_tick_interval(self):
        return self.interval

    def get_trades(self, market_data):
        """Implement in Strategy subclasses."""
        raise NotImplementedError("get_trades must be implemented")

    def add_orders_uuid(self, trades, columns=["date", "symbol"]):
        """Adds uuid column in trades after grouping by given columns.

        :param: columns: List[str], for each group of given columns, one single uuid will be attached
        """
        g = trades.groupby(columns)

        # number of unique class
        ngroups = g.ngroups

        # generate the uuid
        uuids = np.array([str(uuid.uuid4()) for _ in range(ngroups)])

        # map the group number to uuid
        trades["uuid"] = uuids[g.ngroup()]

    def filter_uuids_not_matching_count(self, trades):
        """Filters out trades with uuids not matching the self.exact_number_of_order_per_uuid count"""
        self.add_orders_uuid(trades, self.columns_for_uuid)

        trades.loc[
            :, "exact_number_of_orders_per_uuid"
        ] = self.exact_number_of_orders_per_uuid

        return trades[
            trades["uuid"].map(trades["uuid"].value_counts())
            == trades["exact_number_of_orders_per_uuid"]
        ]

    def live_orders(self, market_data, tick_time):
        """Returns list of QuantplayExchangeOrder tuples
        which are to be executed immediately.
        """
        try:
            trades = self.get_trades(market_data)

            trades.loc[:, "tag"] = self.strategy_tag
            trades.loc[:, "strategy_type"] = self.strategy_type

            trades = trades[trades.date == tick_time]
            trades = trades.to_dict("records")
            orders = QuantplayExchangeOrder.get_exchange_orders(
                trades,
                order_type=self.order_type,
                exchange=self.exchange_to_trade_on,
                execution_algo=self.execution_algo,
            )

            return orders
        except Exception as e:
            raise StrategyInvocationException(
                f"Failed to invoke strategy {self.strategy_tag}"
            ) from e

    def add_derivative_symbols(self, mode='backtest'):
        opt_derivative_symbols, fut_derivative_symbols = set(), set()

        equity_symbols = []
        if "EQ" in self.stream_symbols_by_security_type:
            equity_symbols = [
                MarketConstants.INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP.get(symbol, symbol)
                for symbol in self.stream_symbols_by_security_type["EQ"]
            ]

        if hasattr(self, "option_nearest_expiry_offset"):
            depth = None
            after_date = None
            before_date = None
            if hasattr(self, "option_chain_depth"):
                depth = self.option_chain_depth
            if hasattr(self, "backtest_after_date"):
                after_date = self.backtest_after_date
            if hasattr(self, "backtest_before_date"):
                before_date = self.backtest_before_date
            if mode == 'prod':
                after_date = str(datetime.datetime.now().date())
                before_date = str(datetime.datetime.now().date())

            self.opt_expiry_data = market.get_nearest_expiry_data(
                symbols=equity_symbols,
                days_offset=self.option_nearest_expiry_offset,
                security_type="OPT",
                depth=depth,
                after_date=after_date,
                before_date=before_date,
            )
            opt_expiry_data_records = self.opt_expiry_data.to_dict("records")
            for record in opt_expiry_data_records:
                (symbol, expiry_date, strike_gap, lowest_strike, highest_strike,) = (
                    record["symbol"],
                    record["expiry_date"],
                    int(record["strike_gap"]),
                    int(record["lowest_strike_price"]),
                    int(record["highest_strike_price"]),
                )
                expiry_date = datetime.date.fromisoformat(expiry_date)
                formatted_expiry_date = market.format_expiry_date(
                    expiry_date, security_type="OPT"
                )
                ce_symbols = [
                    symbol + formatted_expiry_date + str(strike) + "CE"
                    for strike in range(lowest_strike, highest_strike, strike_gap)
                ]
                pe_symbols = [
                    symbol + formatted_expiry_date + str(strike) + "PE"
                    for strike in range(lowest_strike, highest_strike, strike_gap)
                ]
                opt_derivative_symbols.update(ce_symbols + pe_symbols)

            self.stream_symbols_by_security_type["OPT"] = list(opt_derivative_symbols)


        if hasattr(self, "future_nearest_expiry_offset"):
            self.fut_expiry_data = market.get_nearest_expiry_data(
                symbols=equity_symbols,
                days_offset=self.future_nearest_expiry_offset,
                security_type="FUT",
            )

            fut_expiry_data_records = self.fut_expiry_data.to_dict("records")
            for record in fut_expiry_data_records:
                symbol, expiry_date = record["symbol"], record["expiry_date"]
                expiry_date = datetime.date.fromisoformat(expiry_date)
                formatted_expiry_date = market.format_expiry_date(
                    expiry_date, security_type="FUT"
                )
                fut_derivative_symbols.add(symbol + formatted_expiry_date + "FUT")

            self.stream_symbols_by_security_type["FUT"] = list(fut_derivative_symbols)


    def add_expiry(self, trades, security_type=None):
        if security_type == "OPT":
            return market.add_expiry(
                    trades,
                    security_type=security_type,
                    days_offset=self.option_nearest_expiry_offset,
                )

        if security_type == "FUT":
            return market.add_expiry(
                    trades,
                    security_type=security_type,
                    days_offset=self.future_nearest_expiry_offset,
                )

    def backtest(self, **kwargs):
        if hasattr(self, "backtest_after_date"):
            kwargs["after"] = self.backtest_after_date
        if hasattr(self, "backtest_before_date"):
            kwargs["before"] = self.backtest_before_date
        market_data = market.data(
            symbols_by_security_type=self.stream_symbols_by_security_type,
            interval=self.interval,
            **kwargs,
        )
        trades = self.get_trades(market_data)
        print("Got {} trades from strategy".format(len(trades)))
        del market_data
        trades.loc[:, "tag"] = self.strategy_tag
        trades.loc[:, "strategy_type"] = self.strategy_type
        trades.loc[:, "entry_time"] = trades.date
        trades.loc[:, "exchange"] = self.exchange_to_trade_on

        holding_days = 0 if (self.strategy_type == "intraday") else self.holding_days

        backtesting.validate_trades(
            trades, columns_to_validate=["transaction_type", "strategy_type", "tag"]
        )

        security_types = list(trades.security_type.unique())
        symbols_by_security_type = {}
        for security_type in security_types:
            symbols_by_security_type[security_type] = list(
                trades[trades.security_type == security_type].tradingsymbol.unique()
            )

        trades.loc[:, "date_only"] = pd.to_datetime(trades.date.dt.date)

        backtesting.load_data(symbols_by_security_type, self.interval)
        if hasattr(self, "exit_time"):
            trades = backtesting.add_time(
                trades,
                column_time_tuple=("exit_time", self.exit_time),
                holding_days=holding_days,
            )
        backtesting.validate_trades(trades, columns_to_validate=["exit_time"])


        results, trades_res = backtesting.evaluate_performance(
            trades, self.interval, tag=self.strategy_tag, args=kwargs
        )

        return results, trades_res



class Market:

    TIMINGS = {
        # In (hour, minute) format
        "NSE": {
            "start": (9, 15),
            "end": (15, 30),
        },
        "NFO": {
            "start": (9, 15),
            "end": (15, 30),
        },
    }

    EXCHANGE_MAPPINGS = {
        "NFO": {"EQ": "NSE", "FUT": "NFO", "OPT": "NFO"},
        "NSE": {"EQ": "NSE", "FUT": "NFO", "OPT": "NFO"},
    }

    INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP = {
        "NIFTY 50": "NIFTY",
        "NIFTY FIN SERVICE": "FINNIFTY",
        "NIFTY BANK": "BANKNIFTY",
    }

    DERIVATIVE_SYMBOL_TO_INDEX_SYMBOL_MAP = {
        "NIFTY": "NIFTY 50",
        "FINNIFTY": "NIFTY FIN SERVICE",
        "BANKNIFTY": "NIFTY BANK",
    }



from quantplay.brokerage.generics.broker import Broker
import pickle
import codecs
import time
from collections import defaultdict
from datetime import timedelta, datetime
from retrying import retry
import pandas as pd
from typing import List, Tuple
from quantplay.utils.data_utils import DataUtils
from quantplay.config.qplay_config import QplayConfig
from quantplay.model.exchange.instrument import QuantplayInstrument
from quantplay.model.exchange.tick import QuantplayTick
from quantplay.model.exchange.order import (
    QuantplayExchangeOrder,
    QuantplayExchangeResponseType,
    QuantplayExchangeResponse,
)
from quantplay.service import market
from alive_progress import alive_bar
from kiteconnect import KiteTicker
from quantplay.utils.constant import Constants

class DatasetPath:
    NSE_EQ = "/NSE_EQ/"
    NSE_OPT = "/NSE_OPT/"
    NSE_FUT = "/NSE_FUT/"


class ZerodhaServiceConstants:
    zerodha_token = "ZERODHA_TOKEN"
    # TODO rename it to kite_object
    zerodha_wrapper = "zerodha_wrapper"
    default_api_type = "default"


class ZBroker(Broker):
    """Zerodha Broker"""

    def __init__(self, tick_intervals, live_trade):
        kite_str = QplayConfig.get_value(ZerodhaServiceConstants.zerodha_wrapper)
        self.kite = pickle.loads(codecs.decode(kite_str.encode(), "base64"))
        self.kws = KiteTicker(self.kite.api_key, self.kite.access_token)
        try:
            self.kite.orders()
        except Exception as e:
            raise Exception(
                "Looks like token has expired. Please regenerate it through 'quantplay broker generate-token'"
            ) from e

        super(ZBroker, self).__init__(tick_intervals, live_trade)

        self.populate_instruments()

    @retry(
        wait_exponential_multiplier=3000,
        wait_exponential_max=10000,
        stop_max_attempt_number=3,
    )
    def populate_instruments(self):
        """Calls zerodha's populate_instruments API and populates instrument mappings."""
        instruments = self.kite.instruments()
        instruments = list(
            map(
                lambda z_instrument: QuantplayInstrument.from_zerodha_instrument(
                    z_instrument
                ),
                instruments,
            )
        )

        Broker.populate_instruments(self, instruments)

    def add_strategy_symbols(self, symbols, exchange):
        super(ZBroker, self).add_strategy_symbols(symbols, exchange)
        if self.live_trade:
            Constants.logger.info("assigning callbacks")
            self.initiate_live_trade_callbacks()

    def get_all_child_orders(self):
        orders_data = self.kite.orders()
        trigger_orders = [
            QuantplayExchangeOrder.from_zerodha_order(data, is_child_order=True)
            for data in orders_data
            if data["status"] == "TRIGGER PENDING" and "tag" in data
        ]
        return trigger_orders

    def place_new_order(self, quantplay_order):
        data = quantplay_order.to_zerodha_order()
        Constants.logger.info("Placing new order through Zerodha {}".format(data))

        try:
            order_id = self.kite.place_order(
                variety=data["variety"],
                tradingsymbol=data["tradingsymbol"],
                exchange=data["exchange"],
                transaction_type=data["transaction_type"],
                quantity=int(abs(data["quantity"])),
                order_type=data["order_type"],
                disclosed_quantity=data["disclosed_quantity"],
                price=data["price"],
                trigger_price=data["trigger_price"],
                product=data["product"],
                stoploss=data["stoploss"],
                tag=data["tag"],
                squareoff=data["squareoff"],
            )
        except Exception as e:
            exception_message = "Order placement failed for [{}] with error [{}]".format(
                quantplay_order, str(e)
            )
            raise Exception(exception_message) from e

        quantplay_order.order_placement_time = datetime.now()
        quantplay_order.order_id = order_id
        return order_id, quantplay_order

    def modify_existing_order(self, quantplay_order):
        data = quantplay_order.to_zerodha_order()
        order_id = quantplay_order.order_id

        try:
            response = self.kite.modify_order(
                order_id=order_id,
                variety=data["variety"],
                price=data["price"],
                trigger_price=data["trigger_price"],
                disclosed_quantity=data["disclosed_quantity"],
                order_type=data["order_type"],
            )

            Constants.logger.info(f"Successfully modified order {response}")
        except Exception as e:
            exception_message = (
                "OrderModificationFailed for {} failed with exception {}".format(
                    order_id, e
                )
            )
            raise Exception(exception_message) from e

        quantplay_order.order_placement_time = datetime.now()
        return quantplay_order

    def stream_new_symbols(self, instrument_ids):
        self.streaming_instruments = list(
            set(self.streaming_instruments + instrument_ids)
        )
        self.kws.subscribe(self.streaming_instruments)
        self.kws.set_mode(self.kws.MODE_FULL, self.streaming_instruments)

    def connect(self):
        self.kws.connect()

    def initiate_live_trade_callbacks(self):
        self.kws.on_connect = self.on_connect
        self.kws.on_ticks = self.on_ticks
        self.kws.on_order_update = self.on_order_update

    def on_connect(self, kws, response):
        """Callback on successfull connect"""
        kws.subscribe(self.streaming_instruments)
        kws.set_mode(kws.MODE_FULL, self.streaming_instruments)

    def on_ticks(self, kws, ticks):
        """Callback on live ticks"""
        quantplay_ticks = list(
            map(
                lambda t: QuantplayTick.from_zerodha_tick(
                    t,
                    self.instrument_id_to_symbol_map,
                    self.instrument_id_to_exchange_map,
                ),
                ticks,
            )
        )
        super(ZBroker, self).market_feed(quantplay_ticks)

    def on_order_update(self, kws, data):
        """Callback on order update"""
        order_id, status, status_message = (
            data["order_id"],
            data["status"],
            data["status_message"],
        )
        quantity, filled_quantity, pending_quantity = (
            data["quantity"],
            data["filled_quantity"],
            data["pending_quantity"],
        )
        average_price = data["average_price"]

        log_message = (
            f"order_id {order_id} status {status} status_message {status_message}"
        )

        Constants.logger.info(f"order_update {log_message} full_data {data}")

        last_processed_order = super(ZBroker, self).get_processed_quantplay_order_by_id(
            order_id
        )
        in_progress_order = super(ZBroker, self).get_in_progress_quantplay_order_by_id(
            order_id
        )

        responses: List[Tuple[QuantplayExchangeResponseType, int, float]] = []
        quantity = 0
        if last_processed_order:
            quantity = filled_quantity - last_processed_order.filled_quantity

        if status == "COMPLETE":
            if not last_processed_order:
                if not in_progress_order:
                    Constants.logger.info(
                        f"[DROPPING_COMPLETE_RESPONSE] order_id {order_id} not in memory"
                    )
                    return
                else:
                    responses.append(
                        (
                            QuantplayExchangeResponseType.NEW_ORDER_CONFIRM,
                            0,
                            0.0,
                        )
                    )
            responses.append(
                (
                    QuantplayExchangeResponseType.TRADE_CONFIRM,
                    quantity,
                    average_price,
                )
            )
        elif status == "UPDATE":
            if not last_processed_order:
                if not in_progress_order:
                    Constants.logger.info(f"[DROPPING_UPDATE_RESPONSE] order_id {order_id} not in memory")
                    return
                elif not status_message:
                    responses.append(
                        (
                            QuantplayExchangeResponseType.NEW_ORDER_CONFIRM,
                            0,
                            0.0,
                        )
                    )
                    responses.append(
                        (
                            QuantplayExchangeResponseType.TRADE_CONFIRM,
                            quantity,
                            0.0,
                        )
                    )
                else:
                    responses.append(
                        (QuantplayExchangeResponseType.NEW_ORDER_REJECT, 0, 0.0)
                    )
            else:
                if not in_progress_order:
                    responses.append(
                        (
                            QuantplayExchangeResponseType.TRADE_CONFIRM,
                            quantity,
                            0.0,
                        )
                    )
                elif not status_message:
                    responses.append(
                        (
                            QuantplayExchangeResponseType.MOD_ORDER_CONFIRM,
                            0,
                            0.0,
                        )
                    )
                    responses.append(
                        (
                            QuantplayExchangeResponseType.TRADE_CONFIRM,
                            quantity,
                            0.0,
                        )
                    )
                else:
                    responses.append(
                        (QuantplayExchangeResponseType.MOD_ORDER_REJECT, 0, 0.0)
                    )
        elif status == "OPEN":
            if not last_processed_order:
                if not in_progress_order:
                    Constants.logger.info(f"[DROPPING_OPEN_RESPONSE] order_id {order_id} not in memory")
                    return
                else:
                    responses.append(
                        (
                            QuantplayExchangeResponseType.NEW_ORDER_CONFIRM,
                            0,
                            0.0,
                        )
                    )
                    responses.append(
                        (
                            QuantplayExchangeResponseType.TRADE_CONFIRM,
                            quantity,
                            0.0,
                        )
                    )
            else:
                Constants.logger.info(f"[DROPPING_OPEN_RESPONSE] order_id {order_id}")
                return
        elif status == "REJECTED":
            if not last_processed_order:
                if not in_progress_order:
                    Constants.logger.info(
                        f"[DROPPING_REJECTED_RESPONSE] order_id {order_id} not in memory"
                    )
                    return
                else:
                    responses.append(
                        (QuantplayExchangeResponseType.NEW_ORDER_REJECT, 0, 0.0)
                    )
            else:
                if not in_progress_order:
                    Constants.logger.info(f"[UNEXPECTED_REJECTED_RESPONSE] order_id {order_id}")
                    return
                else:
                    responses.append(
                        (QuantplayExchangeResponseType.MOD_ORDER_REJECT, 0, 0.0)
                    )
        elif status == "CANCELLED":
            responses.append((QuantplayExchangeResponseType.CANCEL_ORDER_CONFIRM, 0, 0.0))
        elif status == "TRIGGER PENDING":
            if not last_processed_order:
                if not in_progress_order:
                    Constants.logger.info(
                        f"[DROPPING_TRIGGER_PENDING_RESPONSE] order_id {order_id} not in memory"
                    )
                    return
                else:
                    responses.append(
                        (QuantplayExchangeResponseType.TRIGGER_CONFIRM, 0, 0.0)
                    )
            else:
                Constants.logger.info(
                    f"[DROPPING_TRIGGER_PENDING_RESPONSE] order_id {order_id} already in last_processed map"
                )
                return
        else:
            Constants.logger.info(f"[ZERODHA_UNKNOWN_STATUS] {log_message} full_data {data}")
            return

        for response_type, __quantity, avg_price in responses:
            order_resp = QuantplayExchangeResponse(
                order_id, response_type, __quantity, avg_price, status_message
            )
            super(ZBroker, self).on_order_response(order_resp)

    def is_valid_live_tick(self, tick):
        today_day = datetime.now()
        if tick.exchange in ["NSE", "NFO"]:
            tick_time = tick.last_trade_time
            if (
                (tick_time.hour < 9)
                or (tick_time.hour > 15)
                or (tick_time.day != today_day.day)
                or (tick_time.hour != today_day.hour)
                or (tick_time.minute != today_day.minute)
            ):
                # print("Incorrect Trade time {}".format(tick))
                return False
            elif tick_time.hour == 9 and tick_time.minute < 15:
                self.pre_market_volume[tick.symbol] = tick.volume_traded
                return False
            elif tick_time.hour == 15 and tick_time.minute > 28:
                Constants.logger.info("Tick after 15:20 stock {} tick {}".format(tick.symbol, tick))
                return False

            return True

    @retry(
        wait_exponential_multiplier=1000,
        wait_exponential_max=10000,
        stop_max_attempt_number=8,
    )
    def get_historical_data_from_kite(
        self, instrument, start_date, end_date, interval, continuous=False
    ):
        return self.kite.historical_data(
            instrument, start_date, end_date, interval, continuous=continuous
        )

    def get_historical_data(self, instrument, start_date, end_date):
        data_by_interval = dict()

        for interval in self.tick_intervals:
            if interval == "5minute":
                days_diff = 100
            elif interval == "minute":
                days_diff = 60
            elif interval == "day":
                days_diff = 2000
            else:
                raise Exception("interval {} not whitelisted".format(interval))

            time_diff = timedelta(days=days_diff)

            data = []
            while end_date > (start_date + time_diff):
                Constants.logger.info(
                    "querying data from [%s] [%s], instrument [%s] interval [%s]"
                    % (start_date, start_date + time_diff, instrument, interval)
                )
                temp_data = self.get_historical_data_from_kite(
                    instrument,
                    start_date,
                    start_date + time_diff,
                    interval,
                    continuous=False,
                )

                start_date += time_diff

                if temp_data is not None and len(temp_data) > 0:
                    data = data + temp_data

            Constants.logger.info(
                "querying data from [%s] [%s] instrument [%s] interval [%s]"
                % (start_date, end_date, instrument, interval)
            )
            temp_data = self.get_historical_data_from_kite(
                instrument, start_date, end_date, interval, continuous=False
            )
            data = data + temp_data
            Constants.logger.info("Interval {} Total size {}".format(interval, len(data)))

            df = pd.DataFrame(data)
            if len(df) > 0:
                symbol = self.instrument_id_to_symbol_map[instrument]
                df.loc[:, "symbol"] = symbol
                df.loc[:, "date"] = df["date"].dt.tz_localize(None)
                df.loc[:, "date"] = pd.to_datetime(df.date)

            data_by_interval[interval] = df

        return data_by_interval

    def get_equity_instruments(self, main_instruments):
        stock_list = market.symbols()
        main_instruments = [
            a for a in main_instruments if a["tradingsymbol"] in stock_list
        ]

        indices_instruments = [
            a for a in self.get_instruments() if a["segment"] == "INDICES"
        ]
        main_instruments += indices_instruments

        return main_instruments

    @retry(
        wait_exponential_multiplier=3000,
        wait_exponential_max=10000,
        stop_max_attempt_number=3,
    )
    def filter_instruments(self, exchange, segment, name=None):
        main_instruments = self.kite.instruments()

        main_instruments = [a for a in main_instruments if a["exchange"] == exchange]
        main_instruments = [a for a in main_instruments if a["segment"] == segment]

        if name is not None:
            main_instruments = [a for a in main_instruments if name in a["name"]]

        return main_instruments



import configparser
from os.path import expanduser
import os

# TODO: Remove qplay_config from qplay-cli and instead, import this in the qplay-cli


class QplayConfig:
    """Contains methods for writing and reading Quantplay config."""

    config_path = "{}/.quantplay".format(expanduser("~"))

    def __init__(self):
        pass

    @staticmethod
    def get_credentials():
        isExist = os.path.exists(QplayConfig.config_path)
        if not isExist:
            # Create a new directory because it does not exist
            os.makedirs(QplayConfig.config_path)

        credentials = configparser.ConfigParser()
        credentials.read("{}/config".format(QplayConfig.config_path))

        return credentials

    @staticmethod
    def get_config():
        return QplayConfig.get_credentials()

    def get_value(key):
        return QplayConfig.get_config()["DEFAULT"][key]

    @staticmethod
    def save_credentials(access_token):
        credentials = QplayConfig.get_credentials()
        credentials["DEFAULT"]["access_token"] = access_token
        with open("{}/config".format(QplayConfig.config_path), "w") as configfile:
            credentials.write(configfile)

    @staticmethod
    def save_config(key, value):
        credentials = QplayConfig.get_credentials()
        credentials["DEFAULT"][key] = value
        with open("{}/config".format(QplayConfig.config_path), "w") as configfile:
            credentials.write(configfile)



import seaborn as sns
from matplotlib import pyplot as plt
import pandas as pd
import copy


class VisualReport:
    @staticmethod
    def display_profit_report(trades, groupby="M", file_prefix=""):
        """Saves profit report and displays it.

        :param: groupby: M for monthly, Y for yearly, W for weekly
        """

        groupby_filter_to_str_mapper = {"M": "Month", "Y": "Year", "W": "Week"}

        file_path = "/tmp/{}profit_return_grouped_by_{}.png".format(
            file_prefix, groupby_filter_to_str_mapper[groupby]
        )
        df = (
            trades.groupby([pd.Grouper(key="order_timestamp", freq=groupby)])["profit"]
            .sum()
            .reset_index()
        )
        df.loc[:, "order_timestamp"] = df.order_timestamp.dt.date.astype(str)
        sns.set(rc={"figure.figsize": (11.7, 8.27)})
        sns.set_theme(style="whitegrid")
        ax = sns.barplot(x="order_timestamp", y="profit", data=df).set(
            title="Profit by {}".format(groupby_filter_to_str_mapper[groupby])
        )
        plt.xticks(rotation=90)
        plt.savefig(file_path)
        plt.show()

    @staticmethod
    def display_balance_report(trades, file_prefix):
        file_path = "/tmp/{}balance.png".format(file_prefix)
        trades.loc[:, "balance"] = trades.sort_values(["order_timestamp"]).profit.cumsum()
        trades.plot(kind="line", x="date", y="balance", color="blue")
        plt.title("Balance Progression")
        plt.savefig(file_path)
        plt.show()



import unittest

from quantplay.strategy.musk import Musk


class QuantplayStrategyValidation(unittest.TestCase):
    def test_musk_execution(self):
        musk = Musk()
        musk.validate()
        result, trades = musk.backtest(
            before="2022-02-27 00:00:00", after="2022-01-01 00:00:00"
        )

        actual_trades = trades[trades.date_only == "2022-02-24"]
        actual_trades = actual_trades.to_dict("records")

        put_trades = [a for a in actual_trades if "PE" in a["tradingsymbol"]]
        call_trades = [a for a in actual_trades if "CE" in a["tradingsymbol"]]

        assert len(put_trades) == 2
        assert len(call_trades) == 2

        nifty_put_trade = [a for a in put_trades if a["symbol"] == "NIFTY 50"][0]
        nifty_call_trade = [a for a in call_trades if a["symbol"] == "NIFTY 50"][0]
        bank_nifty_put_trade = [a for a in put_trades if a["symbol"] == "NIFTY BANK"][0]
        bank_nifty_call_trade = [a for a in call_trades if a["symbol"] == "NIFTY BANK"][0]

        self.assertEqual(nifty_put_trade["tradingsymbol"], "NIFTY22FEB16650PE")
        self.assertEqual(nifty_put_trade["entry_price"], 101.3)
        self.assertEqual(nifty_put_trade["close_price"], 151.95)
        self.assertEqual(nifty_put_trade["quantity"], 100)

        self.assertEqual(nifty_call_trade["tradingsymbol"], "NIFTY22FEB16650CE")
        self.assertEqual(nifty_call_trade["entry_price"], 93.45)
        self.assertEqual(nifty_call_trade["close_price"], 0.20)
        self.assertEqual(nifty_call_trade["quantity"], 100)

        self.assertEqual(bank_nifty_put_trade["tradingsymbol"], "BANKNIFTY22FEB36400PE")
        self.assertEqual(bank_nifty_put_trade["entry_price"], 269.10)
        self.assertEqual(bank_nifty_put_trade["close_price"], 484.40)
        self.assertEqual(bank_nifty_put_trade["quantity"], 50)

        self.assertEqual(bank_nifty_call_trade["tradingsymbol"], "BANKNIFTY22FEB36400CE")
        self.assertEqual(bank_nifty_call_trade["entry_price"], 249.10)
        self.assertEqual(bank_nifty_call_trade["close_price"], 0.40)
        self.assertEqual(bank_nifty_call_trade["quantity"], 50)


if __name__ == "__main__":
    unittest.main()



from quantplay.utils.constant import (
    Product,
    OrderType,
    OrderVariety,
    Order,
    ExchangeName,
    StrategyType,
)
import pandas as pd
import copy
from enum import Enum
from typing import NamedTuple
from quantplay.utils.constant import Constants


class QuantplayTransactionStatus(Enum):
    IN_PROGRESS = 1  # Current QuantplayExchangeOrder sent for execution is in progress and no response is received yet.
    PROCESSED = 2  # A response has been received for the sent QuantplayExchangeOrder


class QuantplayExchangeResponseType(Enum):

    NEW_ORDER_CONFIRM = 1
    NEW_ORDER_REJECT = 2
    MOD_ORDER_CONFIRM = 3
    MOD_ORDER_REJECT = 4
    TRADE_CONFIRM = 5
    OMS_REJECT = 6
    RMS_REJECT = 7
    CANCEL_ORDER_CONFIRM = 8
    TRIGGER_CONFIRM = 9


class QuantplayExchangeOrderStatus(Enum):
    OPEN = 1
    COMPLETE = 2
    REJECTED = 3
    CANCELLED = 4

    @classmethod
    def from_response_type(cls, response_type: QuantplayExchangeResponseType):

        if response_type in [
            QuantplayExchangeResponseType.NEW_ORDER_CONFIRM,
            QuantplayExchangeResponseType.MOD_ORDER_CONFIRM,
            QuantplayExchangeResponseType.MOD_ORDER_REJECT,
            QuantplayExchangeResponseType.TRADE_CONFIRM,
        ]:
            return cls.OPEN
        elif response_type in [
            QuantplayExchangeResponseType.NEW_ORDER_REJECT,
            QuantplayExchangeResponseType.OMS_REJECT,
            QuantplayExchangeResponseType.RMS_REJECT,
        ]:
            return cls.REJECTED
        elif response_type == QuantplayExchangeResponseType.CANCEL_ORDER_CONFIRM:
            return cls.CANCELLED
        else:
            raise Exception(f"Invalid response_type {response_type}")


class QuantplayExchangeOrder:
    @staticmethod
    def get_exchange_orders(
        trades,
        order_type=OrderType.limit,
        exchange=ExchangeName.nse,
        execution_algo=None,
    ):
        all_trades = []
        for trade in trades:
            all_trades.append(
                QuantplayExchangeOrder(
                    trade,
                    order_type=order_type,
                    exchange=exchange,
                    execution_algo=execution_algo,
                )
            )
        return all_trades

    def initialize_basic_fields(self):
        self.is_child_order = False
        self.order_id = None
        self.modifications = -1  # Indicating it's a new order
        self.order_placement_time = None
        self.transaction_status: QuantplayTransactionStatus = None
        self.order_status: QuantplayExchangeOrderStatus = None
        self.filled_quantity = 0
        self.average_price = 0.0

    def __init__(
        self,
        trade,
        order_type=OrderType.limit,
        exchange=ExchangeName.nse,
        execution_algo=None,
    ):

        self.initialize_basic_fields()
        self.execution_algo = execution_algo
        self.order_type = str(order_type)
        self.exchange = exchange

        if "tag" not in trade:
            raise Exception(
                "Tag should be present in trade but not found. Trade: {}".format(trade)
            )
        if "quantity" not in trade:
            raise Exception(
                "Quantity should be present in trade but not found. Trade: {}".format(
                    trade
                )
            )
        if "tradingsymbol" not in trade:
            raise Exception(
                "tradingsymbol should be present in trade but not found. Trade: {}".format(
                    trade
                )
            )
        if "strategy_type" not in trade:
            raise Exception(
                "strategy_type should be present in trade but not found. Trade: {}".format(
                    trade
                )
            )

        self.tag = str(trade["tag"])
        self.quantity = abs(int(trade["quantity"]))
        self.strategy_type = trade["strategy_type"]
        self.tradingsymbol = str(trade["tradingsymbol"])
        self.transaction_type = trade["transaction_type"]

        # TODO CRITICAL
        self.price = None
        self.trigger_price = None
        self.order_timestamp = None

        if "stoploss" in trade:
            self.stoploss = float(trade["stoploss"])
        else:
            self.stoploss = None

        if "price" in trade:
            self.price = Constants.round_to_tick(float(trade["price"]))

        if "squareoff" in trade:
            self.squareoff = float(trade["squareoff"])
        else:
            self.squareoff = None

        if "trigger_price" in trade:
            self.trigger_price = Constants.round_to_tick(float(trade["trigger_price"]))

        if "order_timestamp" in trade:
            self.order_timestamp = trade["order_timestamp"]

        if "lot_size" in trade and not pd.isna(trade["lot_size"]):
            self.lot_size = trade["lot_size"]

    def get_child_order(self):
        """Returns the child order for this QuantplayExchangeOrder"""
        if self.is_child_order:
            raise Exception(f"[ALREADY_CHILD_ORDER] order is already a child. {self}")

        child_order = copy.deepcopy(self)

        child_order.initialize_basic_fields()
        child_order.is_child_order = True
        child_order.order_type = OrderType.sl

        if self.transaction_type == "SELL":
            child_order.transaction_type = "BUY"
            child_order.trigger_price = child_order.price * (1 + child_order.stoploss)
            child_order.price = child_order.trigger_price * 1.03
        else:
            child_order.transaction_type = "SELL"
            child_order.trigger_price = child_order.price * (1 - child_order.stoploss)
            child_order.price = child_order.trigger_price * 0.97

        child_order.price = Constants.round_to_tick(child_order.price)
        child_order.trigger_price = Constants.round_to_tick(child_order.trigger_price)

        return child_order

    def modify_child_order_for_closing(self):
        self.order_type = OrderType.limit

    def to_zerodha_order(self):
        """Translates QuantplayExchangeOrder to zerodha order"""
        data = {}
        data["variety"] = OrderVariety.regular
        data["tradingsymbol"] = self.tradingsymbol
        data["exchange"] = self.exchange
        data["transaction_type"] = self.transaction_type
        data["quantity"] = self.quantity
        data["order_type"] = self.order_type
        data["disclosed_quantity"] = None
        data["price"] = self.price
        data["trigger_price"] = self.trigger_price

        if self.exchange == ExchangeName.nfo:
            data["product"] = Product.nrml
        if (
            self.exchange == ExchangeName.nse
            and self.strategy_type == StrategyType.intraday
        ):
            data["product"] = Product.mis
        if (
            self.exchange == ExchangeName.nse
            and self.strategy_type == StrategyType.overnight
        ):
            data["product"] = Product.cnc

        data["stoploss"] = self.stoploss
        data["tag"] = self.tag
        data["squareoff"] = self.squareoff

        return data

    @staticmethod
    def from_zerodha_order(zerodha_order_dict, is_child_order=False):
        d = zerodha_order_dict
        trade = {}
        trade["tag"] = d["tag"]
        trade["quantity"] = d["quantity"]
        trade["tradingsymbol"] = d["tradingsymbol"]
        trade["transaction_type"] = d["transaction_type"]
        trade["price"] = d["price"]
        trade["trigger_price"] = d["trigger_price"]
        trade["order_timestamp"] = d["order_timestamp"]
        trade["strategy_type"] = None

        order = QuantplayExchangeOrder(
            trade, order_type=d["order_type"], exchange=d["exchange"], execution_algo=None
        )
        order.order_id = d["order_id"]
        order.is_child_order = is_child_order
        return order

    def set_strategy_attributes(self, strategy_obj):
        self.strategy_type = strategy_obj.strategy_type
        self.execution_algo = strategy_obj.execution_algo

    def __repr__(self):
        return str(self.__dict__)


class QuantplayExchangeResponse(NamedTuple):
    order_id: str
    response_type: QuantplayExchangeResponseType
    quantity: int
    average_price: float
    message: str

    def __repr__(self):
        return f"order_id {self.order_id} response_type {self.response_type} quantity {self.quantity} average_price {self.average_price} message {self.message}"



from quantplay.brokerage.generics.broker import Broker
from smartapi import SmartConnect
from quantplay.config.qplay_config import QplayConfig
from quantplay.model.exchange.instrument import QuantplayInstrument
import pickle
from retrying import retry
import codecs
from datetime import timedelta, datetime
from quantplay.utils.constant import Constants
import pandas as pd
import numpy as np
import getpass
import traceback
from smartapi import SmartWebSocket
from quantplay.model.exchange.order import (
    QuantplayExchangeOrder,
    QuantplayExchangeResponseType,
    QuantplayExchangeResponse,
)
import requests
import json


class AngelBroker(Broker):
    angelone_api_key = "angelone_api_key"
    angelone_api_secret = "angelone_api_secret"
    angelone_client_id = "angelone_client_id"
    angelone_wrapper = "angelone_wrapper"
    angel_refresh_token = "angel_refresh_token"

    def __init__(self, tick_intervals, live_trade):
        try:
            wrapper = QplayConfig.get_value(AngelBroker.angelone_wrapper)
            self.wrapper = pickle.loads(codecs.decode(wrapper.encode(), "base64"))
            self.refreshToken = QplayConfig.get_value(AngelBroker.angel_refresh_token)
            user_profile_response = self.wrapper.getProfile(self.refreshToken)
            if user_profile_response['message'] != "SUCCESS":
                raise Exception("AngelOne Token Expired")
            else:
                print(user_profile_response)
        except Exception as e:
            Constants.logger.error(e)
            self.wrapper = self.generate_token()
            print(self.wrapper.getProfile(self.refreshToken))
        self.refreshToken = QplayConfig.get_value(AngelBroker.angel_refresh_token)
        self.client_id = QplayConfig.get_value(AngelBroker.angelone_client_id)

        self.angelone_ws = SmartWebSocket(self.wrapper.getfeedToken(), self.client_id)
        super(AngelBroker, self).__init__(tick_intervals, live_trade)
        self.populate_instruments()

    def get_all_child_orders(self):
        orders_data = self.wrapper.orderBook()['data']
        if orders_data == None:
            return []
        trigger_orders = [
            QuantplayExchangeOrder.from_zerodha_order(data, is_child_order=True)
            for data in orders_data
            if data["status"] == "TRIGGER PENDING" and "tag" in data
        ]
        return trigger_orders

    @retry(
        wait_exponential_multiplier=1000,
        wait_exponential_max=10000,
        stop_max_attempt_number=8,
    )
    def get_historical_data_from_angelone(self, instrument, start_date, end_date, interval):
        return self.wrapper.getCandleData({
            "exchange": "NSE",
            "symboltoken": instrument,
            "interval": interval,
            "fromdate": start_date,
            "todate": end_date
        })['data']

    def get_historical_data(self, instrument, start_date, end_date):
        data_by_interval = dict()

        interval_map = {
            "minute": "ONE_MINUTE"
        }
        for interval in self.tick_intervals:
            if interval == "5minute":
                days_diff = 100
            elif interval == "minute":
                days_diff = 60
            elif interval == "day":
                days_diff = 2000
            else:
                raise Exception("interval {} not whitelisted".format(interval))

            time_diff = timedelta(days=days_diff)

            data = []
            while end_date > (start_date + time_diff):
                Constants.logger.info(
                    "querying data from [%s] [%s], symbol [%s] interval [%s]"
                    % (start_date, start_date + time_diff, self.instrument_id_to_symbol_map[instrument], interval)
                )
                temp_data = self.get_historical_data_from_angelone(instrument,
                                                                   start_date.strftime('%Y-%m-%d %H:%M'),
                                                                   (start_date + time_diff).strftime('%Y-%m-%d %H:%M'),
                                                                   interval_map[interval])

                start_date += time_diff

                if temp_data is not None and len(temp_data) > 0:
                    data = data + temp_data

            Constants.logger.info(
                "querying data from [%s] [%s] symbol [%s] interval [%s]"
                % (start_date, end_date, self.instrument_id_to_symbol_map[instrument], interval)
            )
            temp_data = self.get_historical_data_from_angelone(instrument,
                                                               start_date.strftime('%Y-%m-%d %H:%M'),
                                                               end_date.strftime('%Y-%m-%d %H:%M'),
                                                               interval_map[interval])
            data = data + temp_data
            Constants.logger.info("Interval {} Total size {}".format(interval, len(data)))

            df = pd.DataFrame(data)
            df.rename(columns={0: 'date', 1: 'open', 2: 'high', 3:'low', 4:'close', 5: 'volume'}, inplace=True)

            if len(df) > 0:
                symbol = self.instrument_id_to_symbol_map[instrument]
                df.loc[:, "symbol"] = symbol
                df.loc[:, 'date'] = pd.to_datetime(df.date)
                df.loc[:, "date"] = df["date"].dt.tz_localize(None)
                df.loc[:, "date"] = pd.to_datetime(df.date)
                df.to_csv('/tmp/angelone_df.csv')

            data_by_interval[interval] = df

        return data_by_interval

    def populate_instruments(self):
        data = requests.get("https://margincalculator.angelbroking.com/OpenAPI_File/files/OpenAPIScripMaster.json")
        inst_data = json.loads(data.content)
        inst_data = pd.DataFrame(inst_data)
        inst_data.loc[:, 'exchange'] = inst_data.exch_seg
        inst_data = inst_data[inst_data.exchange.isin(["NSE", "NFO"])]
        inst_data.loc[:, 'instrument_token'] = inst_data.token.astype(int)
        inst_data.loc[:, 'symbol'] = inst_data['symbol'].str.replace('-EQ','')

        assert set(['OPTSTK', 'OPTIDX', 'FUTSTK', 'FUTIDX']) == set(inst_data[inst_data.exch_seg == "NFO"].instrumenttype.unique())

        inst_data.loc[:, 'segment'] = None
        inst_data.loc[:, 'segment'] = np.where((inst_data.exch_seg == "NFO") & (
                    (inst_data.instrumenttype == "OPTIDX") | (inst_data.instrumenttype == "OPTSTK")),
                                               "NFO-OPT", inst_data.segment)
        inst_data.loc[:, 'segment'] = np.where((inst_data.exch_seg == "NFO") & (
                    (inst_data.instrumenttype == "FUTIDX") | (inst_data.instrumenttype == "FUTSTK")),
                                               "NFO-FUT", inst_data.segment)
        inst_data.loc[:, 'segment'] = np.where(inst_data.exch_seg == "NSE",
                                               "NSE", inst_data.segment)
        inst_data = inst_data[~inst_data.segment.isna()]
        inst_data.loc[:, 'instrument_type'] = np.where(inst_data.segment == "NFO-FUT", "FUT", None)
        inst_data.loc[:, 'instrument_type'] = np.where(inst_data.segment == "NSE", "EQ",
                                                       inst_data.instrument_type)
        inst_data.loc[:, 'instrument_type'] = np.where(
            ((inst_data.segment == "NFO-OPT") & (inst_data.symbol.str[-2:] == "PE")),
            "PE", inst_data.instrument_type)
        inst_data.loc[:, 'instrument_type'] = np.where(
            ((inst_data.segment == "NFO-OPT") & (inst_data.symbol.str[-2:] == "CE")),
            "CE", inst_data.instrument_type)
        inst_data = inst_data.to_dict('records')

        instruments = list(
            map(
                lambda z_instrument: QuantplayInstrument.from_angelone_instrument(
                    z_instrument
                ),
                inst_data,
            )
        )

        Broker.populate_instruments(self, instruments)

    def connect(self):
        self.angelone_ws.connect()

    def initiate_live_trade_callbacks(self):
        self.angelone_ws._on_open = self.on_open
        self.angelone_ws._on_message = self.on_message
        self.angelone_ws._on_error = self.on_error
        self.angelone_ws._on_close = self.on_close

    def on_message(self, ws, message):
        print("Ticks: {}".format(message))

    def on_open(self, ws):
        print("on open")
        stream_data = "&".join(["nse_cm|{}".format(str(x)) for x in self.streaming_instruments])
        self.angelone_ws.subscribe("mw", stream_data)

    def on_error(self, ws, error):
        print(error)

    def on_close(self, ws):
        print("Close")


    def add_strategy_symbols(self, symbols, exchange):
        super(AngelBroker, self).add_strategy_symbols(symbols, exchange)
        if self.live_trade:
            Constants.logger.info("assigning callbacks")
            self.initiate_live_trade_callbacks()

    def configure(self):
        quantplay_config = QplayConfig.get_config()

        print("Enter AngelOne API key:")
        api_key = input()

        print("Enter AngelOne API Secret:")
        api_secret = input()

        print("Enter AngelOne Client ID:")
        client_id = input()

        quantplay_config['DEFAULT'][AngelBroker.angelone_api_key] = api_key
        quantplay_config['DEFAULT'][AngelBroker.angelone_api_secret] = api_secret
        quantplay_config['DEFAULT'][AngelBroker.angelone_client_id] = client_id

        with open('{}/config'.format(QplayConfig.config_path), 'w') as configfile:
            quantplay_config.write(configfile)

    def validate_config(self, quantplay_config):
        if quantplay_config is None:
            return False
        if AngelBroker.angelone_api_key not in quantplay_config['DEFAULT']:
            return False
        if AngelBroker.angelone_api_secret not in quantplay_config["DEFAULT"]:
            return False
        if AngelBroker.angelone_client_id not in quantplay_config["DEFAULT"]:
            return False

        return True

    def generate_token(self):
        quantplay_config = QplayConfig.get_config()

        if not self.validate_config(quantplay_config):
            self.configure()
            quantplay_config = QplayConfig.get_config()
        api_key = quantplay_config['DEFAULT'][AngelBroker.angelone_api_key]
        api_secret = quantplay_config['DEFAULT'][AngelBroker.angelone_api_secret]
        client_id = quantplay_config['DEFAULT'][AngelBroker.angelone_client_id]
        wrapper = SmartConnect(api_key=api_key)

        password = getpass.getpass()
        data = wrapper.generateSession(client_id, password)
        self.refreshToken = data['data']['refreshToken']
        QplayConfig.save_config(AngelBroker.angel_refresh_token, self.refreshToken)

        QplayConfig.save_config("angelone_wrapper", codecs.encode(pickle.dumps(wrapper), "base64").decode())
        return wrapper


from quantplay.services.market import Market
from quantplay.backtest.backtest_trades import Backtesting
from quantplay.reporting.strategy_report import StrategyReport

market = Market()
backtesting = Backtesting(market)
tradelens = StrategyReport



from typing import NamedTuple, List
from datetime import datetime
from collections import defaultdict
from quantplay.utils.constant import Constants

class BookRecord(NamedTuple):
    price: float
    orders: int
    quantity: int

    @classmethod
    def from_zerodha_book_record(cls, zerodha_book_record_dict):
        """Returns BookRecord from book record present in depth
        field of zerodha tick
        """
        d = zerodha_book_record_dict
        return cls(d["price"], d["orders"], d["quantity"])


class OrderBook(NamedTuple):
    sell_side: List[BookRecord]
    buy_side: List[BookRecord]

    @classmethod
    def from_zerodha_depth(cls, zerodha_depth_dict):
        """Creates OrderBook from Zerodha Depth Dict
        received in the tick data
        """
        d = zerodha_depth_dict
        sell_depth = d["sell"]
        buy_depth = d["buy"]

        sell_side_book = [BookRecord.from_zerodha_book_record(r) for r in sell_depth]
        buy_side_book = [BookRecord.from_zerodha_book_record(r) for r in buy_depth]

        return cls(sell_side_book, buy_side_book)


class QuantplayTick(NamedTuple):
    exchange: str
    instrument_id: int
    symbol: str
    tradable: bool
    last_trade_time: datetime
    open_price: float
    high_price: float
    low_price: float
    close_price: float
    oi: int
    oi_day_high: int
    oi_day_low: int
    volume_traded: int
    last_price: float
    average_traded_price: float
    last_traded_quantity: int
    total_buy_quantity: int
    total_sell_quantity: int
    order_book: OrderBook

    @classmethod
    def from_zerodha_tick(
        cls,
        zerodha_tick_dict,
        instrument_id_to_symbol_map,
        instrument_id_to_exchange_map,
    ):
        """Creates QuantplayTick from tick returned by zerodha API"""
        try:
            d = zerodha_tick_dict
            instrument_id = d["instrument_token"]
            symbol = instrument_id_to_symbol_map[instrument_id]
            exchange = instrument_id_to_exchange_map[instrument_id]

            order_book = None
            if "depth" in d:
                order_book = OrderBook.from_zerodha_depth(d["depth"])

            if "last_trade_time" in d:
                last_trade_time = d["last_trade_time"]
            elif "exchange_timestamp" in d:
                last_trade_time = d["exchange_timestamp"]
            else:
                last_trade_time = d["timestamp"]
        except Exception as e:
            Constants.logger.error(f"Failed to convert {zerodha_tick_dict} to Quantplay tick")
            Constants.logger.error(e)

        dd = defaultdict(float, d)

        return cls(
            exchange,
            instrument_id,
            symbol,
            dd["tradable"],
            last_trade_time,
            dd["ohlc"]["open"],
            dd["ohlc"]["high"],
            dd["ohlc"]["low"],
            dd["ohlc"]["close"],
            dd["oi"],
            dd["oi_day_high"],
            dd["oi_day_low"],
            dd["volume_traded"],
            dd["last_price"],
            dd["average_traded_price"],
            dd["last_traded_quantity"],
            dd["total_buy_quantity"],
            dd["total_sell_quantity"],
            order_book,
        )



from quantplay.utils.data_utils import DataUtils
import requests
import json
from quantplay.utils.config_util import QuantplayConfig
import pandas as pd
import numpy as np
from quantplay.exception.exceptions import AccessDeniedException
from quantplay.utils.exchange import Market as MarketConstants
import copy
from datetime import timedelta
from os.path import expanduser
import re
from quantplay.broker.zerodha import Zerodha
from quantplay.broker.angelone import AngelOne
from quantplay.broker.motilal import Motilal
from quantplay.utils.constant import Constants
from datetime import datetime


class Market:
    BASE_URL = "https://7tpcay1yyk.execute-api.ap-south-1.amazonaws.com/prod/"

    GET_SYMBOLS_URL =  "{}get_symbols".format(BASE_URL)
    EXPIRY_DATA_URL = "{}nearest_expiry".format(BASE_URL)
    SECURITY_DATA_URL = "{}get_security_delivery".format(BASE_URL)

    def __init__(self):
        config = QuantplayConfig.get_config()
        self.nse_equity_path = "{}/.quantplay/NSE_EQ/".format(expanduser("~"))
        self.nse_opt_path = "{}/.quantplay/NSE_OPT/".format(expanduser("~"))
        self.nse_fut_path = "{}/.quantplay/NSE_FUT/".format(expanduser("~"))
        self.nse_market_data_path = "{}/.quantplay/NSE_MARKET_DATA/".format(expanduser("~"))
        self.mcx_path = "{}/.quantplay/MCX/".format(expanduser("~"))

        if "DEFAULT" in config and "nse_equity_path" in config["DEFAULT"]:
            self.nse_equity_path = config["DEFAULT"]["nse_equity_path"]
        if "DEFAULT" in config and "nse_opt_path" in config["DEFAULT"]:
            self.nse_opt_path = config["DEFAULT"]["nse_opt_path"]
        if "DEFAULT" in config and "nse_fut_path" in config["DEFAULT"]:
            self.nse_fut_path = config["DEFAULT"]["nse_fut_path"]
        if "DEFAULT" in config and "nse_market_data_path" in config["DEFAULT"]:
            self.nse_market_data_path = config["DEFAULT"]["nse_market_data_path"]
        if "DEFAULT" in config and "mcx_path" in config["DEFAULT"]:
            self.mcx_path = config["DEFAULT"]["mcx_path"]


    def initialize_broker(self, broker_name=None):
        config = QuantplayConfig.get_config()

        if "DEFAULT" in config and "preferred_broker" in config["DEFAULT"] and broker_name is None:
            preferred_broker = config["DEFAULT"]["preferred_broker"]
            Constants.logger.info("Using {} as preferred broker".format(preferred_broker))
            broker_name = preferred_broker

        if broker_name is None:
            raise Exception("please provide broker_name to initialize broker")

        if broker_name == "Zerodha":
            self.broker = Zerodha()
        elif broker_name == "AngelOne":
            self.broker = AngelOne()
        elif broker_name == "Motilal":
            self.broker = Motilal()
        else:
            raise Exception("broker [{}] not supported".format(broker_name))

    def get_nearest_expiry_data(
        self,
        symbols=[],
        days_offset=0,
        security_type=None,
        depth=None,
        after_date=None,
        before_date=None,
    ):
        credentials = QuantplayConfig.get_credentials()
        if "DEFAULT" not in credentials or "access_token" not in credentials["DEFAULT"]:
            raise AccessDeniedException(
                "Access Denied, please signin using [quantplay user signin]"
            )

        access_token = credentials["DEFAULT"]["access_token"]
        input = {
            "access_token": access_token,
            "days_offset": days_offset,
        }

        if symbols:
            input["symbols"] = symbols
        if security_type:
            input["security_type"] = security_type
        if depth is not None:
            input["depth"] = depth

        x = requests.post(Market.EXPIRY_DATA_URL, data=json.dumps(input))
        response = json.loads(x.text)

        if "error" in response and response["error"] == True:
            print(response["message"])
            raise Exception(response["message"])

        data = pd.DataFrame(response["data"])

        # TODO Move after date to API input
        if after_date is not None:
            data = data[data.date >= after_date]
        if before_date is not None:
            data = data[data.date <= before_date]
        return data

    def add_expiry(
        self, data, security_type=None, days_offset=0
    ):
        columns_to_be_added = ["expiry_date", "strike_spread"]

        for column in columns_to_be_added:
            if column in data.columns:
                data = data.drop([column], axis=1)

        symbols = [
            MarketConstants.INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP.get(symbol, symbol)
            for symbol in data.symbol.unique()
        ]


        expiry_data = self.get_nearest_expiry_data(
            symbols=symbols,
            security_type=security_type,
            days_offset=days_offset,
        )

        data.loc[:, "date_only"] = pd.to_datetime(data.date.dt.date)
        expiry_data.loc[:, "expiry_date"] = pd.to_datetime(expiry_data.expiry_date)
        expiry_data.loc[:, "date_only"] = pd.to_datetime(expiry_data.date)

        for (
            idx_derivative_symbol,
            idx_symbol,
        ) in MarketConstants.DERIVATIVE_SYMBOL_TO_INDEX_SYMBOL_MAP.items():
            expiry_data.loc[:, "symbol"] = expiry_data["symbol"].replace(
                [idx_derivative_symbol], idx_symbol
            )

        data = pd.merge(
            data,
            expiry_data[["date_only", "symbol", "expiry_date", "strike_gap"]],
            how="left",
            left_on=["symbol", "date_only"],
            right_on=["symbol", "date_only"],
        )
        return data

    def symbols(self, universe=None, filters={}):
        credentials = QuantplayConfig.get_credentials()
        if "DEFAULT" not in credentials or "access_token" not in credentials["DEFAULT"]:
            raise AccessDeniedException("Access Denied, please signin using [quantplay user signin]")
        input = {}
        if universe != None:
            input["universe"] = universe

        access_token = credentials["DEFAULT"]["access_token"]
        input['access_token'] = access_token
        input['filters'] = filters

        x = requests.post(Market.GET_SYMBOLS_URL, data=json.dumps(input))
        response = json.loads(x.text)

        if "error" in response and response["error"] == True:
            print(response["message"])
            raise Exception(response["message"])

        return response['data']

    def add_security_delivery_data(self, trades):
        credentials = QuantplayConfig.get_credentials()
        if "DEFAULT" not in credentials or "access_token" not in credentials["DEFAULT"]:
            raise AccessDeniedException("Access Denied, please signin using [quantplay user signin]")

        access_token = credentials["DEFAULT"]["access_token"]

        data = copy.deepcopy(trades[['symbol', 'date']])
        data.loc[:, 'date'] = trades.date.dt.date.astype(str)

        request_input = {}
        request_input['access_token'] = access_token
        request_input['data'] = data.to_dict('records')

        x = requests.post(Market.SECURITY_DATA_URL, data=json.dumps(request_input))

        response = json.loads(x.text)
        df = pd.DataFrame(response['data'])
        trades.loc[:, 'yesterdays_delivery_ratio'] = df.yesterdays_delivery_ratio
        trades.loc[:, 'delivery_ratio'] = df.delivery_ratio

    def future_symbol(self, trade_data):
        trades = copy.deepcopy(trade_data)
        trades.loc[:, "tradingsymbol"] = trades["symbol"].replace(
            MarketConstants.INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP
        )

        trades.loc[:, "tradingsymbol"] = trades.tradingsymbol + trades[
            "expiry_date"
        ].dt.strftime("%y").astype(str)

        trades.loc[:, "tradingsymbol"] = trades.tradingsymbol + \
                                         trades["expiry_date"].dt.strftime("%b").str.upper()


        trades.loc[:, "tradingsymbol"] = trades.tradingsymbol + "FUT"
        trades.loc[:, "security_type"] = "FUT"

        return trades

    def option_symbol(self, trade_data, price_column=None, option_type=None):
        trades = copy.deepcopy(trade_data)

        trades.loc[:, "tradingsymbol"] = trades["symbol"].replace(
            MarketConstants.INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP
        )
        trades.loc[:, "tradingsymbol"] = trades.tradingsymbol + trades[
            "expiry_date"
        ].dt.strftime("%y").astype(str)

        trades.loc[:, "month_number"] = (
            trades["expiry_date"].dt.strftime("%m").astype(int).astype(str)
        )
        trades.loc[:, "monthly_option_prefix"] = (
            trades["expiry_date"].dt.strftime("%b").str.upper()
        )

        trades.loc[:, "week_option_prefix"] = np.where(
            trades.month_number.astype(int) >= 10,
            trades.monthly_option_prefix.str[0]
            + trades["expiry_date"].dt.strftime("%d").astype(str),
            trades.month_number + trades["expiry_date"].dt.strftime("%d").astype(str),
        )

        trades.loc[:, "next_expiry"] = trades.expiry_date + pd.DateOffset(days=7)

        trades.loc[:, "tradingsymbol"] = np.where(
            trades.expiry_date.dt.month != trades.next_expiry.dt.month,
            trades.tradingsymbol + trades.monthly_option_prefix,
            trades.tradingsymbol + trades.week_option_prefix,
        )

        trades.loc[:, "tradingsymbol"] = trades.tradingsymbol + trades[
            price_column
        ].astype(str)
        if option_type is not None:
            trades.loc[:, "tradingsymbol"] = trades.tradingsymbol + option_type
        else:
            trades.loc[:, "tradingsymbol"] = trades.tradingsymbol + trades.option_type

        trades.loc[:, "security_type"] = "OPT"

        trades = trades.drop(
            [
                "next_expiry",
                "week_option_prefix",
                "monthly_option_prefix",
                "month_number",
            ],
            axis=1,
        )
        return trades

    @staticmethod
    def filter_contracts_matching_expiry_date(option_data):
        if "expiry_date" not in option_data.columns:
            raise Exception("expiry_date column required in option_data")

        if "symbol" not in option_data.columns:
            raise Exception("symbol column required in option_data")

        return option_data[
            option_data.apply(
                lambda x: x["symbol"].startswith(
                    MarketConstants.INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP[
                        x["equity_symbol"]
                    ]
                    + Market.format_expiry_date(x["expiry_date"], "OPT")
                ),
                axis=1,
            )
        ]

    @staticmethod
    def add_columns_in_option_data(option_data, columns=[]):
        def get_strike_price(x):
            try:
                return int(
                    x["symbol"].split(
                        MarketConstants.INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP[
                            x["equity_symbol"]
                        ]
                        + Market.format_expiry_date(x["expiry_date"], "OPT")
                    )[1][:-2]
                )
            except Exception as e:
                print(x)
                raise e

        columns_supported = ["equity_symbol", "strike_price", "option_type"]
        if not set(columns).issubset(columns_supported):
            raise Exception(
                f"supports only addition of columns {columns_supported}, but received {columns}"
            )

        for col in columns:
            if col == "equity_symbol":
                option_data.loc[:, "equity_symbol"] = option_data.symbol.apply(
                    lambda x: MarketConstants.DERIVATIVE_SYMBOL_TO_INDEX_SYMBOL_MAP[
                        re.findall("([a-zA-Z]*)\d*.*", x)[0]
                    ]
                )
            elif col == "strike_price":
                option_data.loc[:, "strike_price"] = option_data[
                    ["date", "expiry_date", "symbol", "equity_symbol"]
                ].apply(
                    lambda x: get_strike_price(x),
                    axis=1,
                )

            elif col == "option_type":
                option_data.loc[:, "option_type"] = option_data.symbol.apply(
                    lambda x: x[-2:]
                )

    @staticmethod
    def format_expiry_date(expiry_date, security_type):
        formatted_date = expiry_date.strftime("%y")
        month_number, month_name, day = (
            expiry_date.month,
            expiry_date.strftime("%b").upper(),
            expiry_date.strftime("%d"),
        )

        if security_type == "FUT":
            formatted_date += month_name
        elif security_type == "OPT":
            if month_number >= 10:
                week_prefix = month_name[0] + day
            else:
                week_prefix = str(month_number) + day

            next_expiry = expiry_date + timedelta(days=7)
            if next_expiry.month == expiry_date.month:
                formatted_date += week_prefix
            else:
                formatted_date += month_name
        else:
            raise Exception(
                "Invalid security_type {}. Date argument was {}".format(
                    security_type, expiry_date
                )
            )

        return formatted_date

    def equity_data(self, interval=None, symbols=None):
        df = DataUtils.load_data_using_pandas(
            stocks=symbols, interval=interval, path=self.nse_equity_path
        )
        df.loc[:, "security_type"] = "EQ"
        return df

    def mcx_data(self, interval=None, symbols=None):
        df = DataUtils.load_data_using_pandas(
            stocks=symbols, interval=interval, path=self.mcx_path
        )
        return df

    def data(self, interval=None, symbols_by_security_type=None, **kwargs):
        security_type_data_path = {
            "EQ": self.nse_equity_path,
            "FUT": self.nse_fut_path,
            "OPT": self.nse_opt_path,
        }
        dfs = []
        for security_type, symbols in symbols_by_security_type.items():

            print(
                f"Loading {security_type} data for {len(set(symbols))} symbols, interval {interval}"
            )
            df = DataUtils.load_data_using_pandas(
                stocks=symbols,
                interval=interval,
                path=security_type_data_path[security_type],
                ignore_if_not_found=True,
                **kwargs,
            )
            df.loc[:, "security_type"] = security_type
            dfs.append(df)

        return pd.concat(dfs, axis=0)

    def options_data(self, interval=None, symbols=None):
        df = DataUtils.load_data_using_pandas(
            stocks=symbols, interval=interval, path=self.nse_opt_path
        )
        df.loc[:, "security_type"] = "OPT"
        return df

    def future_data(self, interval=None, symbols=None):
        df = DataUtils.load_data_using_pandas(
            stocks=symbols, interval=interval, path=self.nse_fut_path
        )
        df.loc[:, "security_type"] = "FUT"
        return df

    def get_trades(self, data, entry_time):
        return data[
            data.assign(Time=data["date"].dt.time.astype(str))["Time"].str.match(
                entry_time
            )
        ]

    @staticmethod
    def merge_price(
        trades,
        data,
        time=None,
        column_name=None,
        merge_on_columns=["symbol", "date_only"],
    ):
        filter_data = data[
            data.assign(Time=data["date"].dt.time.astype(str))["Time"].str.match(time)
        ]

        filter_data.loc[:, column_name] = filter_data.close
        trades = pd.merge(
            trades,
            filter_data[[column_name] + merge_on_columns],
            how="left",
            left_on=merge_on_columns,
            right_on=merge_on_columns,
        )

        return trades

    @staticmethod
    def add_intraday_metrics(data, entry_time, exit_time):
        data.loc[:, "intraday_high"] = data.high
        data.loc[:, "intraday_low"] = data.low

        condition = (data.date.dt.time <= pd.to_datetime(entry_time).time()) | (
            data.date.dt.time > pd.to_datetime(exit_time).time()
        )
        data.loc[:, "intraday_high"] = np.where(condition, np.nan, data.intraday_high)
        data.loc[:, "intraday_low"] = np.where(condition, np.nan, data.intraday_low)

        data.loc[:, "date_only"] = data.date.dt.date
        data.loc[:, "intraday_high"] = data.groupby(
            ["symbol", "date_only"]
        ).intraday_high.transform(max)
        data.loc[:, "intraday_low"] = data.groupby(
            ["symbol", "date_only"]
        ).intraday_low.transform(min)

    def nearest_expiry(self, symbol=None, security_type=None, days_offset=0):
        symbol = MarketConstants.INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP[symbol]
        data = self.get_nearest_expiry_data(
            symbols=[symbol],
            security_type=security_type,
            days_offset=days_offset,
        )
        today_date = str(datetime.now().date())
        data = data[data.date == today_date]

        data = data.to_dict('records')[0]
        data['expiry_date'] = datetime.strptime(data['expiry_date'], '%Y-%m-%d')

        return data

    def day_candle(self, market_data):
        day_candle = market_data.groupby(["symbol", pd.Grouper(key='date', freq="1d")]).agg({
            "open": "first",
            "high": "max",
            "low": "min",
            "close": "last",
            "volume": "sum"
        }).reset_index()

        return day_candle


import time

import pandas as pd
from retrying import retry

from joblib import Parallel, delayed, parallel_backend
from os import path
import os
from quantplay.utils.constant import Constants, TickInterval, timeit


class DataUtils:
    @staticmethod
    @timeit(MetricName="DataUtils:load_file")
    @retry(
        wait_exponential_multiplier=1000,
        wait_exponential_max=10000,
        stop_max_attempt_number=3,
    )
    def load_file(
        interval=None,
        stock=None,
        path=None,
        args=None,
        columns=None,
        ignore_if_not_found=False,
        **kwargs
    ):
        Constants.logger.info("loading {} from {}".format(stock, path))
        path = path + "/" + stock + ".csv"

        df = None
        try:
            df = pd.read_csv(path, usecols=columns)

            if "after" in kwargs:
                df = df[df.date >= kwargs["after"]]
            if "before" in kwargs:
                df = df[df.date <= kwargs["before"]]

            Constants.logger.info("loaded {}".format(path))
            df = df.reset_index(drop=True)
            return df
            # x = pd.read_csv(path)
        except Exception as e:
            Constants.logger.error(
                "Failed to load {}/{}.csv Got {}".format(interval, stock, e)
            )
            if ignore_if_not_found:
                return pd.DataFrame()
            raise e

    @staticmethod
    @timeit(MetricName="DataUtils:load_data_using_pandas")
    def load_data_using_pandas(
        stocks=None,
        interval=None,
        path=None,
        columns=None,
        args=None,
        is_map=False,
        ignore_if_not_found=False,
        **kwargs
    ):
        start_time = time.time()
        response = {}
        if path == None:
            path = Constants.stock_data_path + interval
        else:
            path = path + interval

        response = Parallel(n_jobs=4)(
            delayed(DataUtils.load_file)(
                interval=interval,
                stock=stock,
                path=path,
                columns=columns,
                args=args,
                ignore_if_not_found=ignore_if_not_found,
                **kwargs
            )
            for stock in stocks
        )

        # CandleUtils.candle_count_check_v2(response)
        print(
            "time taken to load data for {} candles --- {} seconds ---".format(
                interval, time.time() - start_time
            )
        )

        response = pd.concat(response).reset_index(drop=True)
        response.loc[:, "date"] = pd.to_datetime(response.date)
        response.loc[:, "open"] = response.open.astype(float)
        response.loc[:, "high"] = response.high.astype(float)
        response.loc[:, "low"] = response.low.astype(float)
        response.loc[:, "close"] = response.close.astype(float)
        response.loc[:, "volume"] = response.volume.astype(float)

        return response

    @staticmethod
    @timeit(MetricName="DataUtils:midday_data_merge_v3")
    def midday_data_merge(
        day_candle_data,
        candle_data,
        timing_value=None,
        columns_to_retrieve=["close", "high", "volume"],
    ):
        hour, minute = timing_value.split(":")
        hour = int(hour)
        minute = int(minute)

        time_df = candle_data[
            (candle_data.date.dt.hour == hour) & (candle_data.date.dt.minute == minute)
        ]

        assert len(time_df) > 0

        for column in columns_to_retrieve:
            time_df.loc[:, "tick_{}".format(column)] = time_df[column]
        columns_to_retrieve = ["tick_{}".format(a) for a in columns_to_retrieve]

        columns_to_drop = columns_to_retrieve
        columns_to_drop = [a for a in columns_to_drop if a in day_candle_data.columns]
        day_candle_data = day_candle_data.drop(columns_to_drop, axis=1)
        day_candle_data = pd.merge(
            day_candle_data,
            time_df[["symbol", "date_only"] + columns_to_retrieve],
            how="left",
            left_on=["date_only", "symbol"],
            right_on=["date_only", "symbol"],
        )

        return day_candle_data

    @staticmethod
    @timeit(MetricName="DataUtils:midday_data_merge_v2")
    def midday_data_merge_v2(day_candle_data, candle_data, columns_to_retrieve=["close"]):

        for column in columns_to_retrieve:
            candle_data.loc[:, "tick_{}".format(column)] = candle_data[column]
        columns_to_retrieve = ["tick_{}".format(a) for a in columns_to_retrieve]

        columns_to_drop = columns_to_retrieve
        columns_to_drop = [a for a in columns_to_drop if a in day_candle_data.columns]
        day_candle_data = day_candle_data.drop(columns_to_drop, axis=1)
        day_candle_data = pd.merge(
            day_candle_data,
            candle_data[["symbol", "date"] + columns_to_retrieve],
            how="left",
            left_on=["tick_timestamp", "symbol"],
            right_on=["date", "symbol"],
        )

        return day_candle_data

    @staticmethod
    @timeit(MetricName="DataUtils:get_data")
    def get_data(path=None, interval=None, stocks=None):
        return DataUtils.load_data_using_pandas(
            stocks=stocks, interval=interval, path=path
        )


if __name__ == "__main__":
    pass



import unittest

from quantplay.strategy.base import QuantplayAlgorithm


class QuantplayAlgorithmValidation(unittest.TestCase):
    def test_max_len_strategy_tag(self):
        self.assertEqual(QuantplayAlgorithm.MAX_LEN_FOR_STRATEGY_TAG, 8)


if __name__ == "__main__":
    unittest.main()



from quantplay.service import market

market.broker.execute_order(tradingsymbol="SBIN",
                            exchange="NSE",
                            quantity=1,
                            order_type="MARKET",
                            transaction_type="BUY",
                            stoploss=0.03,
                            tag='straddle',
                            product="MIS")



import os

import pandas as pd

dir_name = "/Users/ashok/Downloads/stock-options-data"
file_paths = {}
for path, currentDirectory, files in os.walk(dir_name):
    for file in files:
        if "csv" in os.path.join(path, file):
            file_location = os.path.join(path, file)

            if file not in file_paths:
                file_paths[file] = [file_location]
            else:
                file_paths[file].append(file_location)


for file_name in file_paths:
    li = []

    if "FINNIFTY" not in file_name:
        continue

    for file_loc in file_paths[file_name]:
        df = pd.read_csv(file_loc, index_col=None, header=0)
        li.append(df)

    df = pd.concat(li, axis=0, ignore_index=True)

    df.loc[:, "date"] = df["<date>"] + " " + df["<time>"]
    df.loc[:, "date"] = pd.to_datetime(df.date)

    df = df.drop_duplicates(subset=["date"])

    assert len(df) == len(df.date.unique())

    df.rename(
        columns={
            "<ticker>": "symbol",
            "<open>": "open",
            "<close>": "close",
            "<low>": "low",
            "<high>": "high",
            "<volume>": "volume",
        },
        inplace=True,
    )
    df = df[["date", "open", "high", "low", "close", "volume", "symbol"]]

    print("saving file {}".format(file_name))
    df.sort_values(["date"]).reset_index(drop=True).to_csv(
        "~/.quantplay/NSE_OPT/minute/{}".format(file_name), index=False
    )



from quantplay.service import market

symbols = (
    market.symbols("NSE_INDICES")
    + market.symbols("NSE_ALL_STOCKS")
    + ["INDIA VIX"]
    + market.symbols("FNO_STOCKS")
)
symbols = list(set(symbols))


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


symbol_chunks = list(chunks(symbols, 500))

for symbols in symbol_chunks:

    interval = "day"
    market_data = market.data(
        symbols_by_security_type={"EQ": symbols},
        interval=interval,
    )
    market_data = market_data[market_data.date >= "2021-12-01 00:00:00"]
    market_data = market_data[market_data.date < "2022-01-01 00:00:00"]

    symbols = market_data.symbol.unique()

    for symbol in symbols:
        print("Saving data for {}".format(symbol))
        symbol_data = market_data[market_data.symbol == symbol][
            ["symbol", "date", "open", "high", "low", "close", "volume"]
        ]
        symbol_data.to_csv(
            "~/Documents/QuantplaySampleData/NSE_EQ/{}/{}.csv".format(interval, symbol),
            index=False,
        )



import time
import traceback

from retrying import retry
import getpass
from quantplay.utils.selenium_utils import Selenium
from quantplay.config.qplay_config import QplayConfig
import pyotp

class KiteUtils():

    @staticmethod
    @retry(wait_exponential_multiplier=1000, wait_exponential_max=10000, stop_max_attempt_number=3)
    def get_request_token(kite_api_key=None):
        quantplay_config = QplayConfig.get_config()

        try:
            browser = Selenium.get_browser()

            # TODO api should be fetched from configuration

            kite_url = "https://kite.trade/connect/login?api_key={}&v=3".format(kite_api_key)
            print("Kite Url {}".format(kite_url))
            browser.get(kite_url)
            time.sleep(5)

            username = browser.find_element("xpath", '//*[@id="container"]/div/div/div[2]/form/div[1]/input')
            password = browser.find_element("xpath", '//*[@id="container"]/div/div/div[2]/form/div[2]/input')

            if KiteUtils.zerodha_username not in quantplay_config["DEFAULT"]:
                print("Enter zerodha username:")
                username_value = input()
            else:
                username_value = quantplay_config['DEFAULT'][KiteUtils.zerodha_username]

            if KiteUtils.zerodha_password not in quantplay_config["DEFAULT"]:
                print("Enter zerodha password:")
                password_value = getpass.getpass()
            else:
                password_value = quantplay_config['DEFAULT'][KiteUtils.zerodha_password]

            username.send_keys(username_value)
            password.send_keys(password_value)

            login_attempt = browser.find_element("xpath", '//*[@id="container"]/div/div/div[2]/form/div[4]/button')
            login_attempt.submit()
            time.sleep(5)

            kite_pin = browser.find_element("xpath", '//*[@id="container"]/div/div/div[2]/form/div[2]/input')

            if KiteUtils.zeordha_totp_unique_id not in quantplay_config["DEFAULT"]:
                print("Enter TOTP:")
                totp = input()
            else:
                unique_id = quantplay_config['DEFAULT'][KiteUtils.zeordha_totp_unique_id]
                totp = pyotp.totp(unique_id)
                totp = totp.now()

            kite_pin.send_keys(totp)
            time.sleep(3)

            login_attempt = browser.find_element("xpath", '//*[@id="container"]/div/div/div[2]/form/div[3]/button')
            login_attempt.submit()
            time.sleep(5)

            url = browser.current_url
            print("got kite url {}".format(url))
            request_token = url.split('token=')[1].split('&')[0]

            browser.close()

            return request_token
        except:
            print(traceback.print_exc())
            raise


if __name__ == '__main__':
    k = KiteUtils()
    k.get_request_token()



from quantplay.utils.constant import Constants
from collections import defaultdict
from quantplay.utils.exchange import Market as MarketConstants
from datetime import timedelta
from quantplay.exception.exceptions import QuantplayOrderPlacementException

class Broker():

    def __init__(self):
        self.instrument_id_to_symbol_map = dict()
        self.instrument_id_to_exchange_map = dict()
        self.instrument_id_to_security_type_map = dict()
        self.exchange_symbol_to_instrument_id_map = defaultdict(dict)
        self.order_type_sl = "SL"

    def round_to_tick(self, number):
        return round(number * 20) / 20

    def populate_instruments(self, instruments):
        """Fetches instruments for all exchanges from the broker
        and stores them in the member attributes.
        """
        Constants.logger.info("populating instruments")
        for instrument in instruments:
            exchange, symbol, instrument_id = (
                instrument.exchange,
                instrument.symbol,
                instrument.instrument_id,
            )
            self.instrument_id_to_symbol_map[instrument_id] = symbol
            self.instrument_id_to_exchange_map[instrument_id] = exchange
            self.instrument_id_to_security_type_map[
                instrument_id
            ] = instrument.security_type()
            self.exchange_symbol_to_instrument_id_map[exchange][symbol] = instrument_id

    def execute_order(self, tradingsymbol=None, exchange=None, quantity=None, order_type=None, transaction_type=None,
                      stoploss=None, tag=None, product=None, price=None):
        if price is None:
            price = self.get_ltp(exchange=exchange, tradingsymbol=tradingsymbol)
            trade_price = self.get_ltp(exchange=exchange, tradingsymbol=tradingsymbol)
        try:
            if stoploss != None:
                if transaction_type == "SELL":
                    sl_transaction_type = "BUY"
                    sl_trigger_price = self.round_to_tick(price*(1+stoploss))

                    if exchange == "NFO":
                        price = sl_trigger_price*1.05
                    elif exchange == "NSE":
                        price = sl_trigger_price * 1.01
                    else:
                        raise Exception("{} not supported for trading".format(exchange))

                    sl_price = self.round_to_tick(price)
                elif transaction_type == "BUY":
                    sl_transaction_type = "SELL"
                    sl_trigger_price = self.round_to_tick(price * (1 - stoploss))

                    if exchange == "NFO":
                        price = sl_trigger_price*.95
                    elif exchange == "NSE":
                        price = sl_trigger_price * .99
                    else:
                        raise Exception("{} not supported for trading".format(exchange))

                    sl_price = self.round_to_tick(price)
                else:
                    raise Exception("Invalid transaction_type {}".format(transaction_type))
                stoploss_order_id = self.place_order(tradingsymbol=tradingsymbol,
                                                     exchange=exchange,
                                                     quantity=quantity,
                                                     order_type=self.order_type_sl,
                                                     transaction_type=sl_transaction_type,
                                                     tag=tag,product=product, price=sl_price,
                                                     trigger_price=sl_trigger_price)

                if stoploss_order_id is None:
                    Constants.logger.error(
                        "[ORDER_REJECTED] tradingsymbol {}".format(tradingsymbol))
                    raise QuantplayOrderPlacementException("Order reject for {}".format(tradingsymbol))

            if order_type == "MARKET":
                price = 0

            response = self.place_order(tradingsymbol=tradingsymbol, exchange=exchange, quantity=quantity,
                                        order_type=order_type, transaction_type=transaction_type, tag=tag,
                                        product=product, price=trade_price)
            return response
        except Exception as e:
            raise e

    def option_symbol(self, underlying_symbol, expiry_date, strike_price, type):
        option_symbol = MarketConstants.INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP[underlying_symbol]
        option_symbol += expiry_date.strftime('%y')

        month_number = str(int(expiry_date.strftime("%m")))
        monthly_option_prefix = expiry_date.strftime("%b").upper()

        if int(month_number) >= 10:
            week_option_prefix = monthly_option_prefix[0]
        else:
            week_option_prefix = month_number
        week_option_prefix += expiry_date.strftime("%d")

        next_expiry = expiry_date + timedelta(days=7)

        if next_expiry.month != expiry_date.month:
            option_symbol += monthly_option_prefix
        else:
            option_symbol += week_option_prefix

        option_symbol += str(int(strike_price))
        option_symbol += type

        return option_symbol



from quantplay.strategies.options.intraday.musk import Musk
from quantplay.executor.strategy_executor import UserStrategiesExecutor
strategies = [()]

UserStrategiesExecutor("Zerodha", strategies).start_execution()